#include "color_print.h"

#include <stdio.h>
#include <stdarg.h>

static void apply_modif(TextModif modif)
{
	if(modif == TM_NORMAL)
	{
		printf("\033[0m");
		return;
	}
	if(modif & TM_BOLD)
		printf("\033[1m");
	if(modif & TM_UNDERLINE)
		printf("\033[4m");
	if(modif & TM_CROSSED)
		printf("\033[9m");
	if(modif & TM_HIDE)
		printf("\033[8m");	
}

static void apply_color(ConsolColor color, TextModif modif)
{
	if(color < 0 || color >= NB_CCOLOR)
		return;
	printf("\033[%d%d%sm", (modif & TM_INTENSITY) ? 9 : 3, color, (modif & TM_BRIGHT) ? ";1" : "");
}

int cprintf(TextModif modif, ConsolColor color, const char * format, ...)
{
	int retour;
	va_list ap;
	
	printf("\033[0m");
	
	apply_modif(modif);
	apply_color(color, modif);
	
	va_start(ap, format);
	retour = vprintf(format, ap);
	va_end(ap);
	
	printf("\033[0m");
	
	return retour;
}

int rgbprintf(TextModif modif, unsigned char r, unsigned char g, unsigned char b, const char * format, ...)
{
	int retour;
	va_list ap;
	unsigned char color;
	if(r == g && g == b)
	{
		color = 232 + (r * 24) / 256;
	}
	else
	{
		r = (6 * r) / 256;
		g = (6 * g) / 256;
		b = (6 * b) / 256;
		color = 16 + 36 * r + 6 * g + b;
	}
	
	printf("\033[0m");
	
	apply_modif(modif);
	printf("\033[38;5;%dm", color);
	
	va_start(ap, format);
	retour = vprintf(format, ap);
	va_end(ap);
	
	printf("\033[0m");
	
	return retour;
}