#include "construct_poly.h"

#include <stdio.h>
#include <stdlib.h>

int CPoly_create(CPoly * poly, int degree, const GenPolyF2 * based_poly)
{
	GenPolyF2 base;
	int i = 0;
	
	poly->degree = 0;
	if(! GenPolyF2_copy(&base, based_poly))
		return 0;
	if((poly->seq = (GenPolyF2 *)calloc(degree + 1, sizeof(GenPolyF2))) == NULL)
	{
		GenPolyF2_free(&base);
		return 0;
	}
	for(i = 0; i <= degree; ++i)
	{
		if(! GenPolyF2_create(poly->seq + i, PT_NORMAL, 0))
			goto error_zone;
	}
	poly->degree = degree;
	poly->based_poly = base;
	
	return 1;
	
	error_zone:;
	
	for(--i; i >= 0; --i)
		GenPolyF2_free(poly->seq + i);
	free(poly->seq);
	poly->seq = NULL;
	GenPolyF2_free(&base);
	
	return 0;
}

int CPoly_create_by_tab(CPoly * poly, const GenPolyF2 * tab, int nb_vals, const GenPolyF2 * based_poly)
{
	int i;
	if(! CPoly_create(poly, nb_vals - 1, based_poly))
		return 0;
	for(i = 0; i < nb_vals; ++i)
	{
		if(! CPoly_set_val(poly, tab + i, i))
		{
			CPoly_free(poly);
			return 0;
		}
	}
	return 1;
}

void CPoly_free(CPoly * poly)
{
	int i;
	GenPolyF2_free(&(poly->based_poly));
	if(poly->seq != NULL)
	{
		for(i = 0; i <= poly->degree; ++i)
			GenPolyF2_free(poly->seq + i);
		free(poly->seq);
		poly->seq = NULL;
	}
}

int CPoly_real_degree(const CPoly * poly)
{
	int d;
	
	for(d = poly->degree; d >= 0; --d)
	{
		if(! GenPolyF2_equals_vals(poly->seq + d, 0, -1))
			return d;
	}
	return 0;
}

int CPoly_set_val(CPoly * poly, const GenPolyF2 * val, int id)
{
	GenPolyF2 tmp;
	if(id < 0 || id > poly->degree)
		return 0;
	if(! GenPolyF2_copy(&tmp, val))
		return 0;
	GenPolyF2_free(poly->seq + id);
	*(poly->seq + id) = tmp;
	return 1;
}

const GenPolyF2 * CPoly_get_val(const CPoly * poly, int id)
{
	if(id < 0 || id > poly->degree)
		return NULL;
	return poly->seq + id;
}

int CPoly_copy_val(CPoly * poly, GenPolyF2 * val, int id)
{
	const GenPolyF2 * tmp = NULL;
	if((tmp = CPoly_get_val(poly, id)) == NULL)
		return 0;
	if(! GenPolyF2_copy(val, tmp))
		return 0;
	return 1;
}

int CPoly_add_val(CPoly * poly, const GenPolyF2 * val, int id)
{
	if(id < 0 || id > poly->degree)
		return 0;
	return GenPolyF2_self_add(poly->seq + id, val);
}

int CPoly_copy(CPoly * dest, const CPoly * src)
{
	int degree;
	int i;
	
	degree = CPoly_real_degree(src);
	if(! CPoly_create(dest, degree, &(src->based_poly)))
		return 0;
	for(i = 0; i <= degree; ++i)
	{
		if(! CPoly_set_val(dest, CPoly_get_val(src, i), i))
		{
			CPoly_free(dest);
			return 0;
		}
	}
	return 1;
}

int CPoly_to_tab(const CPoly * poly, GenPolyF2 ** tab, int * nb_vals)
{
	GenPolyF2 * res = NULL;
	int degree;
	int i;
	
	degree = CPoly_real_degree(poly);
	if((res = (GenPolyF2 *)calloc(degree + 1, sizeof(GenPolyF2))) == NULL)
		return 0;
	for(i = 0; i <= degree; ++i)
	{
		if(! GenPolyF2_copy(res + i, CPoly_get_val(poly, i)))
		{
			for(--i; i >= 0; --i)
				GenPolyF2_free(res + i);
			free(res);
			return 0;
		}
	}
	*tab = res;
	if(nb_vals != NULL)
		*nb_vals = degree + 1;
	return 1;
}

int CPoly_to_F2(const CPoly * poly, GenPolyF2 * res)
{
	int degree;
	int i;
	
	degree = CPoly_real_degree(poly);
	if(! GenPolyF2_create(res, PT_NORMAL, degree))
		return 0;
	for(i = 0; i <= degree; ++i)
	{
		if(GenPolyF2_equals_vals(CPoly_get_val(poly, i), 0, -1))
			continue;
		if(! GenPolyF2_equals_vals(CPoly_get_val(poly, i), 1, -1))
		{
			GenPolyF2_free(res);
			return 0;
		}
		if(! GenPolyF2_set_val(res, 1, i))
		{
			GenPolyF2_free(res);
			return 0;
		}
	}
	
	return 1;
}

void CPoly_reverse_vals_order(CPoly * poly)
{
	int i;
	int degree_2;
	int degree;
	GenPolyF2 tmp;
	
	degree_2 = (degree = CPoly_real_degree(poly)) / 2;
	for(i = 0; i <= degree_2; ++i)
	{
		tmp = *(poly->seq + i);
		*(poly->seq + i) = *(poly->seq + (degree - i));
		*(poly->seq + (degree - i)) = tmp;
	}
}

int CPoly_extends(CPoly * poly, int add_vals)
{
	int degree;
	GenPolyF2 * seq = NULL;
	if(add_vals <= 0)
		return 1;
	degree = poly->degree + add_vals;
	if((seq = (GenPolyF2 *)realloc(poly->seq, (degree) * sizeof(GenPolyF2))) == NULL)
		return 0;
	poly->seq = seq;
	for(; poly->degree <= degree; ++(poly->degree))
	{
		if(! GenPolyF2_create(poly->seq + (poly->degree), PT_NORMAL, 0))
			return 0;
	}
	--(poly->degree);
	return 1;
}

/* Operation functions */

int CPoly_mul_val(CPoly * poly, const GenPolyF2 * val)
{
	int i;
	int degree;
	
	degree = CPoly_real_degree(poly);
	for(i = 0; i <= degree; ++i)
	{
		if(! GenPolyF2_self_mul(poly->seq + i, val, &(poly->based_poly)))
			return 0;
	}
	return 1;
}

int CPoly_add(CPoly * res, const CPoly * first, const CPoly * second)
{
	int first_degree;
	int second_degree;
	int degree;
	int i;
	
	if(! GenPolyF2_equals(&(first->based_poly), &(second->based_poly)))
		return 0;
	first_degree = CPoly_real_degree(first);
	second_degree = CPoly_real_degree(second);
	degree = ((first_degree > second_degree) ? first_degree : second_degree);
	if(! CPoly_create(res, degree, &(first->based_poly)))
		return 0;
	for(i = 0; i <= degree; ++i)
	{
		if(i > first_degree)
		{
			if(! CPoly_set_val(res, CPoly_get_val(second, i), i))
				goto error_zone;
			continue;
		}
		if(! CPoly_set_val(res, CPoly_get_val(first, i), i))
			goto error_zone;
		if(i > second_degree)
			continue;
		if(! CPoly_add_val(res, CPoly_get_val(second, i), i))
			goto error_zone;
	}
	return 1;
	
	error_zone:;
	
	CPoly_free(res);
	
	return 0;
}

int CPoly_self_add(CPoly * poly, const CPoly * other)
{
	CPoly tmp;
	if(! CPoly_add(&tmp, poly, other))
		return 0;
	CPoly_free(poly);
	*poly = tmp;
	return 1;
}

int CPoly_mul(CPoly * res, const CPoly * first, const CPoly * second)
{
	int first_degree;
	int second_degree;
	int degree;
	int i, j;
	GenPolyF2 tmp;
	
	if(! GenPolyF2_equals(&(first->based_poly), &(second->based_poly)))
		return 0;
	first_degree = CPoly_real_degree(first);
	second_degree = CPoly_real_degree(second);
	degree = first_degree + second_degree;
	if(! CPoly_create(res, degree, &(first->based_poly)))
		return 0;
	for(i = 0; i <= first_degree; ++i)
	{
		for(j = 0; j <= second_degree; ++j)
		{
			if(! GenPolyF2_mul(&tmp, CPoly_get_val(first, i), CPoly_get_val(second, j), &(first->based_poly)))
			{
				CPoly_free(res);
				return 0;
			}
			if(! GenPolyF2_self_add((GenPolyF2 *)CPoly_get_val(res, i + j), &tmp))
			{
				GenPolyF2_free(&tmp);
				CPoly_free(res);
				return 0;
			}
			GenPolyF2_free(&tmp);
		}
	}
	
	return 1;
}

int CPoly_self_mul(CPoly * poly, const CPoly * other)
{
	CPoly tmp;
	if(! CPoly_mul(&tmp, poly, other))
		return 0;
	CPoly_free(poly);
	*poly = tmp;
	return 1;
}

int CPoly_euclid_div(CPoly * factor, CPoly * remind, const CPoly * first, const CPoly * second)
{
	CPoly _factor;
	CPoly tmp_first;
	CPoly tmp_second;
	int first_degree;
	int second_degree;
	GenPolyF2 tmp_f2;
	int i;
	int last_deg;
	
	if((first_degree = CPoly_real_degree(first)) < (second_degree = CPoly_real_degree(second)))
	{
		if(remind != NULL)
		{
			if(! CPoly_copy(remind, first))
				return 0;
		}
		if(factor != NULL)
		{
			if(! CPoly_create(factor, 0, &(first->based_poly)))
				return 0;
		}
	}
	if(! CPoly_copy(&tmp_first, first))
		return 0;
	if(! CPoly_create(&_factor, first_degree - second_degree, &(first->based_poly)))
	{
		CPoly_free(&tmp_first);
		return 0;
	}
	last_deg = -1;
	while(second_degree <= first_degree && last_deg != first_degree)
	{
		if(! GenPolyF2_invert(&tmp_f2, CPoly_get_val(second, second_degree), &(first->based_poly)))
		{
			CPoly_free(&_factor);
			CPoly_free(&tmp_first);
			return 0;
		}
		if(! GenPolyF2_self_mul(&tmp_f2, CPoly_get_val(&tmp_first, first_degree), &(second->based_poly))
		|| ! CPoly_copy(&tmp_second, second))
			goto error_zone;
		if(! CPoly_mul_val(&tmp_second, &tmp_f2))
			goto error_zone;
		for(i = 0; i <= second_degree; ++i)
		{
			if(! CPoly_add_val(&tmp_first, CPoly_get_val(&tmp_second, i), first_degree - second_degree + i))
			{
				CPoly_free(&tmp_second);
				goto error_zone;
			}
		}
		if(! CPoly_set_val(&_factor, &tmp_f2, first_degree - second_degree))
		{
			CPoly_free(&tmp_second);
			goto error_zone;
		}
		GenPolyF2_free(&tmp_f2);
		CPoly_free(&tmp_second);
		last_deg = first_degree;
		
		first_degree = CPoly_real_degree(&tmp_first);
		/*if(GenPolyF2_equals_vals(CPoly_get_val(&tmp_first, first_degree), 0, -1))
			continue;
		if(! GenPolyF2_invert(&tmp_f2, CPoly_get_val(&tmp_first, first_degree), &(first->based_poly)))
		{
			CPoly_free(&_factor);
			CPoly_free(&tmp_first);
			return 0;
		}
		if(! CPoly_mul_val(&tmp_first, &tmp_f2))
			goto error_zone;
		GenPolyF2_free(&tmp_f2);*/
	}
	if(factor != NULL)
		*factor = _factor;
	else
		CPoly_free(&_factor);
	if(remind != NULL)
		*remind = tmp_first;
	else
		CPoly_free(&tmp_first);
	
	return 1;
	
	error_zone:;
	
	CPoly_free(&_factor);
	CPoly_free(&tmp_first);
	GenPolyF2_free(&tmp_f2);
	return 0;
}

static int is_gcd_base(const CPoly * first, const CPoly * second, void * user_data)
{
	if(CPoly_real_degree(second) != 0)
		return 0;
	return GenPolyF2_equals_vals(CPoly_get_val(second, 0), 0, -1);
}

static int CPoly_xgcd_aux(CPoly * gcd, CPoly * u, CPoly * v, const CPoly * first, const CPoly * second,
	const CPoly * u_i1, const CPoly * v_i1, const CPoly * u_i2, const CPoly * v_i2,
	int (*is_gcd)(const CPoly * first, const CPoly * second, void * user_data), void * user_data)
{
	int first_degree, second_degree;
	int success;
	CPoly remind, factor;
	CPoly u_i, v_i;
	
	if((first_degree = CPoly_real_degree(first)) < (second_degree = CPoly_real_degree(second)))
		return CPoly_xgcd_aux(gcd, u, v, second, first, v_i1, u_i1, v_i2, u_i2, is_gcd, user_data);
	
	if((*is_gcd)(first, second, user_data))
	{
		if(gcd != NULL)
		{
			if(! CPoly_copy(gcd, first))
				return 0;
		}
		if(u != NULL)
		{
			if(! CPoly_copy(u, u_i2))
				return 0;
		}
		if(v != NULL)
		{
			if(! CPoly_copy(v, v_i2))
				return 0;
		}
		return 1;
	}
	
	if(! CPoly_euclid_div(&factor, &remind, first, second))
		return 0;
	
	if(! CPoly_mul(&u_i, &factor, u_i1))
		goto error_zone;
	if(! CPoly_self_add(&u_i, u_i2))
	{
		CPoly_free(&u_i);
		goto error_zone;
	}
	
	if(! CPoly_mul(&v_i, &factor, v_i1))
	{
		CPoly_free(&u_i);
		goto error_zone;
	}
	if(! CPoly_self_add(&v_i, v_i2))
	{
		CPoly_free(&u_i);
		CPoly_free(&v_i);
		goto error_zone;
	}
	
	success = CPoly_xgcd_aux(gcd, u, v, second, &remind, &u_i, &v_i, u_i1, v_i1, is_gcd, user_data);
	
	CPoly_free(&u_i);
	CPoly_free(&v_i);
	CPoly_free(&factor);
	CPoly_free(&remind);
	
	return success;
	
	error_zone:;
	
	CPoly_free(&factor);
	CPoly_free(&remind);
	
	return 0;
}

int CPoly_xgcd(CPoly * gcd, CPoly * u, CPoly * v, const CPoly * first, const CPoly * second,
	int (*is_gcd)(const CPoly * first, const CPoly * second, void * user_data), void * user_data)
{
	CPoly u_0, v_0, u_1, v_1;
	GenPolyF2 c0, c1;
	int success;
	
	if(! GenPolyF2_equals(&(first->based_poly), &(second->based_poly)))
		return 0;
	
	if(! GenPolyF2_create(&c0, PT_NORMAL, 0)
	|| ! GenPolyF2_create_single(&c1, PT_NORMAL, 0))
		return 0;
	if(! CPoly_create(&u_0, 0, &(first->based_poly))
	|| ! CPoly_create(&u_1, 0, &(first->based_poly))
	|| ! CPoly_create(&v_0, 0, &(first->based_poly))
	|| ! CPoly_create(&v_1, 0, &(first->based_poly)))
	{
		GenPolyF2_free(&c0);
		GenPolyF2_free(&c1);
		return 0;
	}
	if(! CPoly_set_val(&u_0, &c1, 0)
	|| ! CPoly_set_val(&u_1, &c0, 0)
	|| ! CPoly_set_val(&v_0, &c0, 0)
	|| ! CPoly_set_val(&v_1, &c1, 0))
	{
		GenPolyF2_free(&c0);
		GenPolyF2_free(&c1);
		CPoly_free(&u_0);
		CPoly_free(&u_1);
		CPoly_free(&v_0);
		CPoly_free(&v_1);
		return 0;
	}
	GenPolyF2_free(&c0);
	GenPolyF2_free(&c1);
	
	success = CPoly_xgcd_aux(gcd, u, v, first, second,
		&u_1, &v_1, &u_0, &v_0,
		(is_gcd == NULL) ? (&is_gcd_base) : (is_gcd), user_data);
	
	CPoly_free(&u_0);
	CPoly_free(&u_1);
	CPoly_free(&v_0);
	CPoly_free(&v_1);
	return success;
}

int CPoly_gcd(CPoly * gcd, CPoly * u, CPoly * v, const CPoly * first, const CPoly * second)
{
	return CPoly_xgcd(gcd, u, v, first, second, NULL, NULL);
}

static int is_gcd_under_degree(const CPoly * first, const CPoly * second, void * user_data)
{
	if(CPoly_real_degree(first) < *((int *)user_data))
		return 1;
	return 0;
}

int CPoly_gcd_under_degree(CPoly * gcd, CPoly * u, CPoly * v, const CPoly * first, const CPoly * second, int degree)
{
	return CPoly_xgcd(gcd, u, v, first, second, &is_gcd_under_degree, &degree);
}

int CPoly_power(CPoly * res, const CPoly * poly, int power)
{
	CPoly exp;
	CPoly tmp;
	
	if(power <= 0)
	{
		GenPolyF2 un;
		if(! GenPolyF2_create_single(&un, PT_NORMAL, 0))
			return 0;
		if(! CPoly_create_by_tab(res, &un, 1, &(poly->based_poly)))
		{
			GenPolyF2_free(&un);
			return 0;
		}
		GenPolyF2_free(&un);
		return 1;
	}
	if(power == 1)
		return CPoly_copy(res, poly);
	if(! CPoly_power(&exp, poly, power / 2))
		return 0;
	if(! CPoly_mul(&tmp, &exp, &exp))
	{
		CPoly_free(&exp);
		return 0;
	}
	CPoly_free(&exp);
	if(power % 2 == 0)
	{
		*res = tmp;
		return 1;
	}
	if(! CPoly_self_mul(&tmp, poly))
	{
		CPoly_free(&tmp);
		return 0;
	}
	*res = tmp;
	return 1;
}

int CPoly_compose(CPoly * res, const CPoly * poly, const CPoly * var)
{
	CPoly tmp_power;
	
	int poly_degree;
	int i;
	
	poly_degree = CPoly_real_degree(poly);
	if(! CPoly_create(res, 0, &(poly->based_poly)))
		return 0;
	for(i = 0; i <= poly_degree; ++i)
	{
		if(GenPolyF2_equals_vals(CPoly_get_val(poly, i), 0, -1))
			continue;
		if(! CPoly_power(&tmp_power, var, i))
			goto error_zone;
		if(! CPoly_mul_val(&tmp_power, CPoly_get_val(poly, i)))
		{
			CPoly_free(&tmp_power);
			goto error_zone;
		}
		if(! CPoly_self_add(res, &tmp_power))
		{
			CPoly_free(&tmp_power);
			goto error_zone;
		}
		CPoly_free(&tmp_power);
	}
	
	return 1;
	
	error_zone:;
	
	CPoly_free(res);
	
	return 0;
}

int CPoly_eval(GenPolyF2 * res, const CPoly * poly, const GenPolyF2 * var)
{
	CPoly tmp;
	CPoly p_res;
	if(! CPoly_create_by_tab(&tmp, var, 1, &(poly->based_poly)))
		return 0;
	if(! CPoly_compose(&p_res, poly, &tmp))
	{
		CPoly_free(&tmp);
		return 0;
	}
	CPoly_free(&tmp);
	if(CPoly_real_degree(&p_res) != 0)
	{
		CPoly_free(&p_res);
		return 0;
	}
	if(! GenPolyF2_copy(res, CPoly_get_val(&p_res, 0)))
	{
		CPoly_free(&p_res);
		return 0;
	}
	CPoly_free(&p_res);
	return 1;
}

/* Other */

static void print_val(const GenPolyF2 * val)
{
	int degree;
	int i;
	unsigned char gval;
	
	degree = GenPolyF2_real_degree(val);
	for(i = degree; i >= 0; --i)
	{
		GenPolyF2_get_val(val, &gval, i);
		putchar('0' + gval);
	}
}

void CPoly_print(const CPoly * poly)
{
	int nb;
	int i;
	int degree;
	
	nb = 0;
	degree = CPoly_real_degree(poly);
	for(i = degree; i >= 0; --i)
	{
		if(GenPolyF2_equals_vals(poly->seq + i, 0, -1))
			continue;
		if(nb != 0)
			printf(" + ");
		++nb;
		if(! GenPolyF2_equals_vals(poly->seq + i, 1, -1) || i == 0)
			print_val(poly->seq + i);
		if(i == 0)
			continue;
		putchar('X');
		if(i == 1)
			continue;
		printf("^%d", i);
	}
	if(nb == 0)
		putchar('0');
	printf(" [");
	print_val(&(poly->based_poly));
	printf("]\n");
}

void CPoly_print_hexa(const CPoly * poly)
{
	int nb;
	int i;
	int degree;
	
	nb = 0;
	degree = CPoly_real_degree(poly);
	for(i = degree; i >= 0; --i)
	{
		if(GenPolyF2_equals_vals(poly->seq + i, 0, -1))
			continue;
		if(nb != 0)
			printf(" + ");
		++nb;
		if(! GenPolyF2_equals_vals(poly->seq + i, 1, -1) || i == 0)
			GenPolyF2_print_hexa(poly->seq + i);
		if(i == 0)
			continue;
		putchar('X');
		if(i == 1)
			continue;
		printf("^%d", i);
	}
	if(nb == 0)
		putchar('0');
	printf(" [");
	GenPolyF2_print_hexa(&(poly->based_poly));
	printf("]\n");
}

void CPoly_print_calcul(const CPoly * poly)
{
	int nb;
	int i;
	int degree;
	
	nb = 0;
	degree = CPoly_real_degree(poly);
	for(i = degree; i >= 0; --i)
	{
		if(GenPolyF2_equals_vals(poly->seq + i, 0, -1))
			continue;
		if(nb != 0)
			printf(" . ");
		++nb;
		if(! GenPolyF2_equals_vals(poly->seq + i, 1, -1) || i == 0)
			print_val(poly->seq + i);
		if(i == 0)
			continue;
		putchar('X');
		if(i == 1)
			continue;
		printf("%d", i);
	}
	if(nb == 0)
		putchar('0');
	putchar('\n');
}

int CPoly_locator(CPoly * res, const GenPolyF2 * based_poly, const GenPolyF2 * syndroms, int nb_syndroms)
{
	CPoly s_x;
	CPoly x2t;
	CPoly l_x;
	GenPolyF2 un;
	GenPolyF2 inv;
	
	if(! CPoly_create_by_tab(&s_x, syndroms, nb_syndroms * 2, based_poly))
		return 0;
	/*printf("S(x)\n");
	CPoly_print_hexa(&s_x);*/
	if(! CPoly_create(&x2t, nb_syndroms * 2, based_poly))
	{
		CPoly_free(&s_x);
		return 0;
	}
	if(! GenPolyF2_create_single(&un, PT_NORMAL, 0))
	{
		CPoly_free(&s_x);
		CPoly_free(&x2t);
		return 0;
	}
	if(! CPoly_set_val(&x2t, &un, nb_syndroms * 2))
	{
		GenPolyF2_free(&un);
		CPoly_free(&s_x);
		CPoly_free(&x2t);
		return 0;
	}
	GenPolyF2_free(&un);
	if(! CPoly_gcd_under_degree(NULL, &l_x, NULL, &s_x, &x2t, nb_syndroms))
	{
		CPoly_free(&s_x);
		CPoly_free(&x2t);
		return 0;
	}
	if(! GenPolyF2_invert(&inv, CPoly_get_val(&l_x, 0), based_poly))
		goto error_zone;
	if(! CPoly_mul_val(&l_x, &inv))
	{
		GenPolyF2_free(&inv);
		goto error_zone;
	}
	GenPolyF2_free(&inv);
	
	*res = l_x;
	CPoly_free(&s_x);
	CPoly_free(&x2t);
	
	return 1;
	
	error_zone:;
	
	CPoly_free(&l_x);
	CPoly_free(&s_x);
	CPoly_free(&x2t);
	
	return 0;
}

int CPoly_formal_derivative(CPoly * res, const CPoly * poly)
{
	int i;
	int degree;
	if(! CPoly_create(res, (degree = CPoly_real_degree(poly)) - 1, &(poly->based_poly)))
		return 0;
	for(i = 0; i < degree; ++i)
	{
		if(i % 2 == 1)
			continue;
		if(! CPoly_set_val(res, CPoly_get_val(poly, i + 1), i))
		{
			CPoly_free(res);
			return 0;
		}
	}
	return 1;
}

int CPoly_to_data(unsigned char ** data, int * size, const CPoly * poly)
{
	unsigned char * res = NULL;
	int degree;
	int i;
	int j;
	unsigned char val;
	degree = CPoly_real_degree(poly);
	if((res = (unsigned char *)calloc(degree + 2, sizeof(unsigned char))) == NULL)
		return 0;
	for(i = 0; i <= degree; ++i)
	{
		for(j = 0; j < 8; ++j)
		{
			GenPolyF2_get_val(CPoly_get_val(poly, degree - i), &val, j);
			*(res + i) |= (val << j);
		}
	}
	if(size != NULL)
		*size = degree + 1;
	*data = res;
	return 1;
}