#ifndef DEF_HEADER_CONSTRUCT_POLY
#define DEF_HEADER_CONSTRUCT_POLY

#include "generic_poly.h"

typedef struct _CPoly CPoly;
struct _CPoly
{
	int degree;
	GenPolyF2 * seq;
	GenPolyF2 based_poly;
};

int CPoly_create(CPoly * poly, int degree, const GenPolyF2 * based_poly);

int CPoly_create_by_tab(CPoly * poly, const GenPolyF2 * tab, int nb_vals, const GenPolyF2 * based_poly);

void CPoly_free(CPoly * poly);

int CPoly_real_degree(const CPoly * poly);

int CPoly_set_val(CPoly * poly, const GenPolyF2 * val, int id);

const GenPolyF2 * CPoly_get_val(const CPoly * poly, int id);

int CPoly_copy_val(CPoly * poly, GenPolyF2 * val, int id);

int CPoly_add_val(CPoly * poly, const GenPolyF2 * val, int id);

int CPoly_copy(CPoly * dest, const CPoly * src);

int CPoly_to_tab(const CPoly * poly, GenPolyF2 ** tab, int * nb_vals);

int CPoly_to_F2(const CPoly * poly, GenPolyF2 * res);

void CPoly_reverse_vals_order(CPoly * poly);

int CPoly_extends(CPoly * poly, int add_vals);

/* Operation functions */

int CPoly_mul_val(CPoly * poly, const GenPolyF2 * val);

int CPoly_add(CPoly * res, const CPoly * first, const CPoly * second);

int CPoly_self_add(CPoly * poly, const CPoly * other);

int CPoly_mul(CPoly * res, const CPoly * first, const CPoly * second);

int CPoly_self_mul(CPoly * poly, const CPoly * other);

int CPoly_euclid_div(CPoly * factor, CPoly * remind, const CPoly * first, const CPoly * second);

int CPoly_xgcd(CPoly * gcd, CPoly * u, CPoly * v, const CPoly * first, const CPoly * second, int (*is_gcd)(const CPoly * first, const CPoly * second, void * user_data), void * user_data);

int CPoly_gcd(CPoly * gcd, CPoly * u, CPoly * v, const CPoly * first, const CPoly * second);

int CPoly_gcd_under_degree(CPoly * gcd, CPoly * u, CPoly * v, const CPoly * first, const CPoly * second, int degree);

int CPoly_power(CPoly * res, const CPoly * poly, int power);

int CPoly_compose(CPoly * res, const CPoly * poly, const CPoly * var);

int CPoly_eval(GenPolyF2 * res, const CPoly * poly, const GenPolyF2 * var);

/* Other */

void CPoly_print(const CPoly * poly);

void CPoly_print_hexa(const CPoly * poly);

void CPoly_print_calcul(const CPoly * poly);

int CPoly_locator(CPoly * res, const GenPolyF2 * based_poly, const GenPolyF2 * syndroms, int nb_syndroms);

int CPoly_formal_derivative(CPoly * res, const CPoly * poly);

int CPoly_to_data(unsigned char ** data, int * size, const CPoly * poly);

#endif