#include "find_primes.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "test_control.h"
#include "matrice.h"

#define ALLOC_BLOC 10

static int a = 0;

/* PrimePoly functions */

int PrimePoly_create(PrimePoly * prime, const PolyIntF2 * poly)
{
	PrimePoly res = NULL;
	
	if((res = (PrimePoly)malloc(sizeof(struct _PrimePoly))) == NULL)
		return 0;
		
	/*if(! PolyIntF2_copy(&(res->poly), poly))
	{
		free(res);
		return 0;
	}*/
	res->poly = *(PolyIntF2 *)poly;
	res->next = NULL;
	
	*prime = res;
	return 1;
}

void PrimePoly_free(PrimePoly * prime)
{
	if(prime == NULL || *prime == NULL)
		return;
	PrimePoly_free(&((*prime)->next));
	PolyIntF2_free(&((*prime)->poly));
	free(*prime);
	*prime = NULL;
}

int PrimePoly_append(PrimePoly * prime, const PolyIntF2 * poly)
{
	if(*prime == NULL)
		return PrimePoly_create(prime, poly);
	if(PolyIntF2_equals(poly, &((*prime)->poly)))
		return 0;
	return PrimePoly_append(&((*prime)->next), poly);
}

void PrimePoly_print(const PrimePoly prime)
{
	if(prime == NULL)
		return;
	PolyIntF2_print(&(prime->poly));
	PrimePoly_print(prime->next);
}

void PrimePoly_print_bin(const PrimePoly prime)
{
	if(prime == NULL)
		return;
	PolyIntF2_print_bin(&(prime->poly));
	PrimePoly_print_bin(prime->next);
}

void PrimePoly_count_aux(const PrimePoly prime, int * count)
{
	if(prime == NULL)
		return;
	++(*count);
	PrimePoly_count_aux(prime->next, count);
}

int PrimePoly_count(const PrimePoly prime)
{
	int count = 0;
	PrimePoly_count_aux(prime, &count);
	return count;
}

int PrimePoly_save(PrimePoly prime, FILE * flow)
{
	GenPolyF2 poly;
	for(; prime != NULL; prime = prime->next)
	{
		if(! GenPolyF2_copy_to_generic(&poly, &(prime->poly), PT_INT))
			return 0;
		if(! GenPolyF2_save(&poly, flow))
		{
			GenPolyF2_free(&poly);
			return 0;
		}
		GenPolyF2_free(&poly);
	}
	return 1;
}

/* PrimePolyList functions */

int PrimePolyList_create(PrimePolyList * list)
{
	list->tab = NULL;
	list->nb = 0;
	list->max_nb = 0;
	
	return 1;
}

void PrimePolyList_free(PrimePolyList * list)
{
	int i;
	if(list->tab != NULL)
	{
		for(i = 0; i < list->nb; ++i)
			PrimePoly_free(list->tab + i);
		free(list->tab);
		list->tab = NULL;
	}
	list->nb = 0;
	list->max_nb = 0;
}

static int PrimePolyList_check_size(PrimePolyList * list, int size)
{
	int new_size = -1;
	PrimePoly * tmp = NULL;
	
	if(size >= list->max_nb)
		new_size = (size * 3) / 2 + ALLOC_BLOC;
	else if(list->nb == list->max_nb)
		new_size = list->max_nb * 2 + ALLOC_BLOC;
	else if(list->nb < list->max_nb / 4)
		new_size = list->max_nb / 2;
	if(new_size == -1)
		return 1;
	
	if((tmp = (PrimePoly *)realloc(list->tab, new_size * sizeof(PrimePoly))) == NULL)
		return 0;
	list->tab = tmp;
	list->max_nb = new_size;
	return 1;
}

int PrimePolyList_append(PrimePolyList * list, const PolyIntF2 * poly)
{
	int i;
	int degree;
	degree = PolyIntF2_real_degree(poly);
	if(! PrimePolyList_check_size(list, degree + 1))
		return 0;
	for(i = list->nb; i <= degree; ++i)
		*(list->tab + i) = NULL;
	list->nb = degree + 1;
	if(! PrimePoly_append(list->tab + degree, poly))
		return 0;
	return 1;
}

void PrimePolyList_print(const PrimePolyList * list)
{
	int i;
	for(i = 0; i < list->nb; ++i)
		PrimePoly_print(*(list->tab + i));
}

void PrimePolyList_print_bin(const PrimePolyList * list)
{
	int i;
	for(i = 0; i < list->nb; ++i)
		PrimePoly_print_bin(*(list->tab + i));
}

int PrimePolyList_save(const PrimePolyList * list, int degree)
{
	FILE * flow;
	if(degree == 20)
		a = 1;
	if(degree < 0 || degree >= list->nb)
		return 0;
	if((flow = open_poly_file(degree, "w+")) == NULL)
		return 0;
	if(! PrimePoly_save(*(list->tab + degree), flow))
	{
		fclose(flow);
		return 0;
	}
	fclose(flow);
	return 1;
}

/* FindingBin functions */

int FindingBin_create(FindingBin * bin, int size)
{
	int i;
	int real_size;
	if((real_size = (size / 8) + (size % 8 != 0)) == 0)
		return 0;
	if((bin->vals = (unsigned char *)malloc(real_size * sizeof(unsigned char))) == NULL)
		return 0;
	bin->size = size;
	for(i = 0; i < real_size; ++i)
		*(bin->vals + i) = 0xff;
	return 1;
}

void FindingBin_free(FindingBin * bin)
{
	if(bin->vals != NULL)
	{
		free(bin->vals);
		bin->vals = NULL;
	}
	bin->size = 0;
}

static int PolyIntF2_to_int(const PolyIntF2 * poly)
{
	int i;
	int degree;
	int res = 0;
	unsigned char val;
	degree = PolyIntF2_real_degree(poly);
	for(i = 0; i <= degree; ++i)
	{
		if(! PolyIntF2_get_val(poly, &val, i))
			return 0;
		if(val)
			res |= ((int)1 << i);
	}
	return res;
}

static int int_to_PolyIntF2(int val, PolyIntF2 * poly)
{
	int i;
	int deg;
	int tmp;
	PolyIntF2 res;
	
	if(val < 0)
		return 0;
	for(tmp = val, i = 0; tmp > 0; ++i, tmp >>= 1);
	deg = i - 1;
	
	if(! PolyIntF2_create(&res, deg))
		return 0;
	for(i = 0; val > 0; ++i, val >>= 1)
	{
		if(! PolyIntF2_set_val(&res, val % 2, i))
		{
			PolyIntF2_free(&res);
			return 0;
		}
	}
	
	*poly = res;
	return 1;
}

void FindingBin_unset(FindingBin * bin, const PolyIntF2 * poly)
{
	int pint;
	pint = PolyIntF2_to_int(poly);
	*(bin->vals + (pint / 8)) &= ~((unsigned char)1 << (pint % 8));
}

/* Real user functions */

static int by_deg = 0;
static int total = 0;

static int MulPoly(FindingBin * bin, PrimePolyList * list, int need_degree, const PolyIntF2 * poly, int id, int prec_deg)
{
	int i;
	PolyIntF2 res;
	PrimePoly cur = NULL;
	int cur_id;
	
	if(need_degree == 0)
	{
		GenPolyF2 poly_test;
		if(GenPolyF2_copy_to_generic(&poly_test, (void *)poly, PT_INT))
		{
			#ifdef USE_ERACHECK
			if(PrimeTest(&poly_test))
			{
				printf("Erreur polynome non premier trouvé comme premier:\n");
				PolyIntF2_print_bin(poly);
			}
			#endif
		}
		else
			printf("Erreur convert\n");
		FindingBin_unset(bin, poly);
		return 1;
	}
	for(i = (prec_deg <= 0) ? 1 : prec_deg; i <= need_degree && i < list->nb; ++i)
	{
		cur_id = 0;
		cur = *(list->tab + i);
		if(prec_deg == i)
		{
			for(; cur_id < id; cur = cur->next, ++cur_id)
				;
		}
		for(; cur != NULL; cur = cur->next, ++cur_id)
		{
			/*printf("-\n");
			PolyIntF2_print_bin(poly);
			PolyIntF2_print_bin(&(cur->poly));*/
			++total;
			++by_deg;
			if(! PolyIntF2_mul(&res, poly, &(cur->poly), NULL))
				return 0;
			if(! MulPoly(bin, list, need_degree - i, &res, cur_id, i))
			{
				PolyIntF2_free(&res);
				return 0;
			}
			PolyIntF2_free(&res);
		}
	}
	return 1;
}

static int FindingBin_add_set(FindingBin * bin, PrimePolyList * list, int degree)
{
	int i;
	PolyIntF2 poly;
	
	for(i = 1 << degree; i < bin->size; ++i)
	{
		if(*(bin->vals + (i / 8)) == 0)
		{
			i += 7;
			continue;
		}
		if(*(bin->vals + (i / 8)) & (1 << (i % 8)))
		{
			if(! int_to_PolyIntF2(i, &poly))
				return 0;
			if(! PrimePolyList_append(list, &poly))
			{
				GenPolyF2 poly_test;
				if(GenPolyF2_copy_to_generic(&poly_test, &poly, PT_INT))
				{
					#ifdef USE_ERACHECK
					if(! PrimeTest(&poly_test))
					{
						printf("Erreur polynome premier trouvé comme produit:\n");
						PolyIntF2_print_bin(&poly);
					}
					#endif
				}
				else
					printf("Erreur convert\n");
				PolyIntF2_free(&poly);
			}
		}
	}
	
	return 1;
}

int FindPolyDegree(PrimePolyList * list, int degree)
{
	FindingBin bin;
	PolyIntF2 start;
	if(degree < 0)
		return 0;
	if(list->nb >= degree + 1)
		return 1;
	if(degree == 0)
	{
		if(! PolyIntF2_create_by_vals(&start, 1, -1))
			return 0;
		return PrimePolyList_append(list, &start);
	}
	if(! FindPolyDegree(list, degree - 1))
		return 0;
	if(! FindingBin_create(&bin, 1 << (degree + 1)))
		return 0;
	if(! PolyIntF2_create_by_vals(&start, 1, -1))
	{
		FindingBin_free(&bin);
		return 0;
	}
	by_deg = 0;
	if(! MulPoly(&bin, list, degree, &start, 0, 0)
	|| ! FindingBin_add_set(&bin, list, degree))
	{
		PolyIntF2_free(&start);
		FindingBin_free(&bin);
		return 0;
	}
	PolyIntF2_free(&start);
	FindingBin_free(&bin);
	if(! PrimePolyList_save(list, degree))
		fprintf(stderr, "Erreur sauvegarde %d\n", degree);
	/*printf("local degree %d => %d\n", degree, by_deg);*/
	/*printf("degree %d = %d\n", degree, total);*/
	printf("degree %d\n", degree);
	return 1;
}

void PrimePolyList_print_count(const PrimePolyList * list)
{
	int i;
	int somme = 0;
	int count;
	
	for(i = 0; i < list->nb; ++i)
	{
		somme += (count = PrimePoly_count(*(list->tab + i)));
		printf("%d of degree %d\n", count, i);
	}
	printf("Total %d\n", somme);
}

int int_is_prime(int val)
{
	int max;
	int i;
	max = sqrt(val) + 1;
	for(i = 2; i <= max; ++i)
	{
		if(val % i == 0)
			return 0;
	}
	return 1;
}

static int PrimeTest_aux(const GenPolyF2 * poly, const GenPolyF2 * start)
{
	int degree;
	int i;
	int success;
	GenPolyF2 res;
	GenPolyF2 last;
	GenPolyF2 gcd;
	const GenPolyF2 * poly_arg = NULL;
	
	if(! GenPolyF2_gcd(&gcd, NULL, NULL, poly, start))
		return 0;
	if(! GenPolyF2_equals_vals(&gcd, 1, -1))
	{
		GenPolyF2_free(&gcd);
		return 0;
	}
	GenPolyF2_free(&gcd);
	degree = GenPolyF2_real_degree(poly);
	for(i = 0; i < degree; ++i)
	{
		poly_arg = (i == 0) ? start : (&last);
		if(! GenPolyF2_mul(&res, poly_arg, poly_arg, poly))
		{
			if(i != 0)
				GenPolyF2_free(&last);
			return 0;
		}
		if(GenPolyF2_real_degree(&res) <= 0
		|| (i < degree - 1 && GenPolyF2_equals(&res, start)))
		{
			/*GenPolyF2_print_bin(start);
			GenPolyF2_print_bin(&res);
			GenPolyF2_print_bin(poly);*/
			if(i != 0)
				GenPolyF2_free(&last);
			GenPolyF2_free(&res);
			return 0;
		}
		if(i != 0)
			GenPolyF2_free(&last);
		last = res;
	}
	success = GenPolyF2_equals(&last, start);
	/*printf("Res: %d\n", success);
	printf("poly: ");
	GenPolyF2_print_bin(poly);
	printf("start: ");
	GenPolyF2_print_bin(start);
	printf("last: ");
	GenPolyF2_print_bin(&last);*/
	GenPolyF2_free(&last);
	return success;
}

int PrimeTest(const GenPolyF2 * poly)
{
	int degree;
	unsigned char val;
	GenPolyF2 start;
	int success = 1;
	int i;
	GenPolyF2 * factors = NULL;
	int nb_factors = 0;
	GenPolyF2 x_1;
	GenPolyF2 tmp;
	
	if((degree = GenPolyF2_real_degree(poly)) <= 0)
	{
		if(! GenPolyF2_get_val(poly, &val, 0))
			return 0;
		return val;
	}
	if(degree == 1)
		return 1;
	if(! GenPolyF2_get_val(poly, &val, 0))
		return 0;
	if(val == 0)
		return 0;
	if(! GenPolyF2_create_by_vals(&x_1, PT_NORMAL, 1, 1, -1))
		return 0;
	if(! GenPolyF2_euclid_div(NULL, &tmp, poly, &x_1))
	{
		GenPolyF2_free(&x_1);
		return 0;
	}
	GenPolyF2_free(&x_1);
	if(GenPolyF2_equals_vals(&tmp, 0, -1))
	{
		GenPolyF2_free(&tmp);
		return 0;
	}
	GenPolyF2_free(&tmp);
		
	if(! GenPolyF2_create_by_vals(&start, PT_NORMAL, 0, 1, -1))
		return 0;
	success = PrimeTest_aux(poly, &start);
	GenPolyF2_free(&start);
	
	if(! success)
		return 0;
	if(int_is_prime(degree))
		return 1;
		
	if(! Berlekamp(poly, &factors, &nb_factors))
		return 0;
	success = GenPolyF2_equals(poly, factors);
	for(i = 0; i < nb_factors; ++i)
		GenPolyF2_free(factors + i);
	free(factors);
	return success;
}

int PrimitiveTest(const GenPolyF2 * poly)
{
	GenPolyF2 racine;
	GenPolyF2 tmp;
	int degree;
	int nb_check;
	int success;
	int i;
	
	if(! PrimeTest(poly))
		return 0;
	if((degree = GenPolyF2_real_degree(poly)) <= 0)
		return 0;
	nb_check = 1 << degree;
	if(! GenPolyF2_create_single(&racine, PT_NORMAL, 1))
		return 0;
	if(! GenPolyF2_copy(&tmp, &racine))
	{
		GenPolyF2_free(&racine);
		return 0;
	}
	for(i = 1; i < nb_check - 1; ++i)
	{
		if(GenPolyF2_equals_vals(&tmp, 1, -1)
		|| ! GenPolyF2_self_mul(&tmp, &racine, poly))
		{
			GenPolyF2_free(&racine);
			GenPolyF2_free(&tmp);
			return 0;
		}
	}
	success = GenPolyF2_equals_vals(&tmp, 1, -1);
	GenPolyF2_free(&racine);
	GenPolyF2_free(&tmp);
	
	return success;
}

int randint(int a, int b)
{
	if(a == b)
		return a;
	return (rand() % (b + 1 - a)) + a;
}

static void gen_aux(GenPolyF2 * poly)
{
	GenPolyF2 * factors = NULL;
	int nb_factors = 0;
	int i;
	if(Berlekamp(poly, &factors, &nb_factors))
	{
		for(i = 0; i < nb_factors; ++i)
		{
			if(PrimeTest(factors + i) && GenPolyF2_add_prime(factors + i))
			{
				printf("Sub-Irreductible (%d) : ", GenPolyF2_real_degree(factors + i));
				GenPolyF2_print_bin(factors + i);
			}
			else if(nb_factors > 1)
				gen_aux(factors + i);
		
			GenPolyF2_free(factors + i);
		}
		free(factors);
	}
}

int PrimeGenerator(int degree_min, int degree_max)
{
	GenPolyF2 poly;
	int rand_deg;
	
	TC_init();
	while(TC_continue())
	{
		rand_deg = randint(degree_min, degree_max);
		if(! int_is_prime(rand_deg))
		{
			if(rand() % rand_deg != 0)
				continue;
		}
		if(! GenPolyF2_gen_random(&poly, rand_deg))
			return 0;
		printf("test : (%d) ", GenPolyF2_real_degree(&poly));
		GenPolyF2_print_bin(&poly);
		if(PrimeTest(&poly))
		{
			if(! GenPolyF2_add_prime(&poly))
				goto error_zone;
			printf("Irreductible\n");
		}
		else
			gen_aux(&poly);
		GenPolyF2_free(&poly);
	}
	TC_quit();
	return 1;
	
	error_zone:;
	
	GenPolyF2_free(&poly);
	TC_quit();
	return 0;
}

static int gen_b_aux(GenPolyF2 * poly)
{
	GenPolyF2 * factors = NULL;
	int nb_factors = 0;
	int i;
	
	if(Berlekamp(poly, &factors, &nb_factors))
	{
		if(nb_factors == 1)
		{
			int success = 1;
			if(GenPolyF2_equals(poly, factors) && (success = GenPolyF2_add_prime(factors)))
			{
				printf("Irreductible (%d) : ", GenPolyF2_real_degree(factors));
				GenPolyF2_print_bin(factors);
			}
			GenPolyF2_free(factors);
			free(factors);
			return success;
		}
		for(i = 0; i < nb_factors; ++i)
		{
			if(! gen_b_aux(factors + i))
				goto error_zone;
			GenPolyF2_free(factors + i);
		}
		free(factors);
	}
	return 1;
	
	error_zone:;
	
	for(; i >= 0; --i)
		GenPolyF2_free(factors + i);
	free(factors);
	
	return 0;
}

int PrimeGeneratorBerlekamp(int degree_min, int degree_max)
{
	GenPolyF2 poly;
	int rand_degre;
	
	TC_init();
	while(TC_continue())
	{
		rand_degre = randint(degree_min, degree_max);
		if(! GenPolyF2_gen_random(&poly, rand_degre))
		{
			TC_quit();
			return 0;
		}
		printf("split : (%d) ", GenPolyF2_real_degree(&poly));
		GenPolyF2_print_bin(&poly);
		if(! gen_b_aux(&poly))
			goto error_zone;
		GenPolyF2_free(&poly);
	}
	TC_quit();
	return 1;
	
	error_zone:;
	
	GenPolyF2_free(&poly);
	TC_quit();
	return 0;
}

int add_tab(GenPolyF2 * gcd, GenPolyF2 ** factors, int * nb_factors)
{
	int k;
	int d1, d2;
	int indice;
	GenPolyF2 tmp;
	GenPolyF2 ngcd;
	
	for(k = 0; k < *nb_factors; ++k)
	{
		if(GenPolyF2_equals(*factors + k, gcd))
		{
			GenPolyF2_free(gcd);
			return 1;
		}
	}
	*(*factors + *nb_factors) = *gcd;
	++(*nb_factors);
	for(k = 0; k < *nb_factors - 1; ++k)
	{
		if(! GenPolyF2_gcd(&ngcd, NULL, NULL, *factors + (*nb_factors - 1), *factors + k))
			return 0;
		if(GenPolyF2_equals_vals(&ngcd, 1, -1))
		{
			GenPolyF2_free(&ngcd);
			continue;
		}
		d1 = GenPolyF2_real_degree(*factors + (*nb_factors - 1));
		d2 = GenPolyF2_real_degree(*factors + k);
		if(d1 > d2)
			indice = (*nb_factors - 1);
		else
			indice = k;
		GenPolyF2_euclid_div(&tmp, NULL, *factors + indice, &ngcd);
		if(GenPolyF2_equals_vals(&tmp, 1, -1))
		{
			GenPolyF2_free(&tmp);
			GenPolyF2_free(&ngcd);
			continue;
		}
		GenPolyF2_free(*factors + indice);
		*(*factors + indice) = tmp;
		GenPolyF2_free(&ngcd);
	}
	for(k = 0; k < *nb_factors - 1; ++k)
	{
		if(GenPolyF2_equals(*factors + k, *factors + (*nb_factors - 1)))
		{
			GenPolyF2_free(*factors + (*nb_factors - 1));
			--(*nb_factors);
			return 1;
		}
	}
	return 1;
}

int Berlekamp(const GenPolyF2 * poly, GenPolyF2 ** factors, int * nb_factors)
{
	GenPolyF2 derivee;
	GenPolyF2 gcd;
	MatriceF2 mat;
	GenPolyF2 * bases = NULL;
	int nb_bases = 0;
	int i;
	int j;
	
	if(GenPolyF2_real_degree(poly) <= 1)
	{
		if((*factors = (GenPolyF2 *)malloc(sizeof(GenPolyF2))) == NULL)
			return 0;
		if(! GenPolyF2_copy(*factors, poly))
		{
			free(*factors);
			return 0;
		}
		*nb_factors = 1;
		return 1;
	}
	if(! GenPolyF2_differentiate(&derivee, poly))
		return 0;
	if(GenPolyF2_equals_vals(&derivee, 0, -1))
	{
		GenPolyF2_free(&derivee);
		fprintf(stderr, "Polynôme avec racines multiples\n");
		return 0;
	}
	/*GenPolyF2_print(poly);
	GenPolyF2_print(&derivee);*/
	if(! GenPolyF2_gcd(&gcd, NULL, NULL, poly, &derivee))
	{
		GenPolyF2_free(&derivee);
		return 0;
	}
	/*GenPolyF2_print(&gcd);*/
	if(! GenPolyF2_equals_vals(&gcd, 1, -1))
	{
		/*printf("gcd\n");*/
		GenPolyF2_free(&derivee);
		if((*factors = (GenPolyF2 *)malloc(2 * sizeof(GenPolyF2))) == NULL)
			return 0;
		**factors = gcd;
		GenPolyF2_euclid_div(*factors + 1, NULL, poly, &gcd);
		*nb_factors = 2;
		return 1;
	}
	GenPolyF2_free(&gcd);
	GenPolyF2_free(&derivee);
	if(! MatriceF2_Berlekamp(&mat, poly))
		return 0;
	/*MatriceF2_print(&mat);*/
	if(! MatriceF2_get_ker_id(&mat, &bases, &nb_bases))
	{
		MatriceF2_free(&mat);
		return 0;
	}
	MatriceF2_free(&mat);
	if((*factors = (GenPolyF2 *)malloc(nb_bases * sizeof(GenPolyF2))) == NULL)
		return 0;
	*nb_factors = 0;
	/*printf("%d bases\n", nb_bases);*/
	for(i = 0; i < nb_bases; ++i)
	{
		/*GenPolyF2_print_bin(bases + i);*/
		if(GenPolyF2_real_degree(bases + i) <= 0)
			continue;
		for(j = 0; j < 2; ++j)
		{
			if(*nb_factors >= nb_bases)
				break;
			GenPolyF2_add_val(bases + i, j, 0);
			if(! GenPolyF2_gcd(&gcd, NULL, NULL, poly, bases + i))
				goto error_zone;
			if(GenPolyF2_real_degree(&gcd) > 0)
			{
				if(! add_tab(&gcd, factors, nb_factors))
					goto error_zone;
			}
			else
				GenPolyF2_free(&gcd);
		}
		if(*nb_factors >= nb_bases)
			break;
	}
	for(i = 0; i < nb_bases; ++i)
		GenPolyF2_free(bases + i);
	free(bases);
	if(*nb_factors == 0)
	{
		if(! GenPolyF2_copy(*factors, poly))
		{
			free(*factors);
			return 0;
		}
		*nb_factors = 1;
		return 1;
	}
	return 1;
	
	error_zone:;
	free(*factors);
	*factors = NULL;
	*nb_factors = 0;
	for(i = 0; i < nb_bases; ++i)
	{
		/*GenPolyF2_print_bin(bases + i);*/
		GenPolyF2_free(bases + i);
	}
	free(bases);
	return 0;
}

/* Save and Load functions */

FILE * open_poly_file(int degree, const char * mode)
{
	char * path = NULL;
	FILE * file = NULL;

	if((path = (char *)malloc((strlen("media/f2_primes/.poly") + 25) * sizeof(char))) == NULL)
		return NULL;
	sprintf(path, "media/f2_primes/%d.poly", degree);
	if((file = fopen(path, mode)) == NULL)
	{
		free(path);
		return NULL;
	}
	free(path);
	return file;
}

int GenPolyF2_load_random(GenPolyF2 * poly, int degree)
{
	long file_size;
	long select;
	int poly_size;
	FILE * poly_file = NULL;
	if((poly_size = size_degree(degree)) < 1)
		return 0;
	
	if((poly_file = open_poly_file(degree, "r")) == NULL)
		return 0;
	
	fseek(poly_file, 0, SEEK_END);
	file_size = ftell(poly_file);
	
	select = rand() % (file_size / poly_size);
	rewind(poly_file);
	fseek(poly_file, select * poly_size, SEEK_SET);
	/*printf("%d %ld %ld\n", poly_size, file_size, select * poly_size);*/
	if(! GenPolyF2_load(poly, degree, poly_file))
	{
		fclose(poly_file);
		return 0;
	}
	fclose(poly_file);
	return 1;
}

int GenPolyF2_add_prime(GenPolyF2 * poly)
{
	FILE * poly_file = NULL;
	int degree;
	GenPolyF2 cmp;
	
	if((degree = GenPolyF2_real_degree(poly)) < 0)
		return 0;
	if((poly_file = open_poly_file(degree, "r")) == NULL)
		goto ajout_zone;
	while(GenPolyF2_load(&cmp, degree, poly_file))
	{
		if(GenPolyF2_equals(poly, &cmp))
		{
			fclose(poly_file);
			GenPolyF2_free(&cmp);
			return 1;
		}
		GenPolyF2_free(&cmp);
	}
	fclose(poly_file);
	
	ajout_zone:;
	if((poly_file = open_poly_file(degree, "a+")) == NULL)
		return 0;
	if(! GenPolyF2_save(poly, poly_file))
	{
		fclose(poly_file);
		return 0;
	}
	fclose(poly_file);
	return 1;
}