#ifndef DEF_HEADER_FIND_PRIMES
#define DEF_HEADER_FIND_PRIMES

#include "poly_int.h"
#include "generic_poly.h"

#include <stdio.h>

typedef struct _PrimePoly * PrimePoly;
struct _PrimePoly
{
	PolyIntF2 poly;
	PrimePoly next;
};

typedef struct _PrimePolyList PrimePolyList;
struct _PrimePolyList
{
	PrimePoly * tab;
	int nb;
	int max_nb;
};

typedef struct _FindingBin FindingBin;
struct _FindingBin
{
	unsigned char * vals;
	int size;
};

/* PrimePoly functions */

int PrimePoly_create(PrimePoly * prime, const PolyIntF2 * poly);

void PrimePoly_free(PrimePoly * prime);

int PrimePoly_append(PrimePoly * prime, const PolyIntF2 * poly);

void PrimePoly_print(const PrimePoly prime);

void PrimePoly_print_bin(const PrimePoly prime);

int PrimePoly_count(const PrimePoly prime);

int PrimePoly_save(PrimePoly prime, FILE * flow);

/* PrimePolyList functions */

int PrimePolyList_create(PrimePolyList * list);

void PrimePolyList_free(PrimePolyList * list);

int PrimePolyList_append(PrimePolyList * list, const PolyIntF2 * poly);

void PrimePolyList_print(const PrimePolyList * list);

void PrimePolyList_print_bin(const PrimePolyList * list);

void PrimePolyList_print_count(const PrimePolyList * list);

int PrimePolyList_save(const PrimePolyList * list, int degree);

/* FindingBin functions */

int FindingBin_create(FindingBin * bin, int size);

void FindingBin_free(FindingBin * bin);

void FindingBin_unset(FindingBin * bin, const PolyIntF2 * poly);

/* Real user functions */

int randint(int a, int b);

int int_is_prime(int val);

int FindPolyDegree(PrimePolyList * list, int degree);

int PrimeTest(const GenPolyF2 * poly);

int PrimitiveTest(const GenPolyF2 * poly);

int PrimeGenerator(int degree_min, int degree_max);

int PrimeGeneratorBerlekamp(int degree_min, int degree_max);

int Berlekamp(const GenPolyF2 * poly, GenPolyF2 ** factors, int * nb_factors);

/* Save and Load functions */

FILE * open_poly_file(int degree, const char * mode);

int GenPolyF2_load_random(GenPolyF2 * poly, int degree);

int GenPolyF2_add_prime(GenPolyF2 * poly);

#endif