#include "generic_poly.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

typedef struct _ConvPoly ConvPoly;
struct _ConvPoly
{
	GenPolyF2 poly;
	unsigned char copied;
	unsigned char is_null;
};

static PolyType larger_type(PolyType first, PolyType second)
{
	return (first > second) ? first : second;
}

static void * get_poly(const GenPolyF2 * poly);

static void * get_conv_tab(ConvPoly * tab, int id)
{
	if((tab + id)->is_null)
		return NULL;
	return get_poly(&((tab + id)->poly));
}

static void free_conv_tab(ConvPoly * tab, int size)
{
	int i;
	for(i = 0; i < size; ++i)
	{
		if((tab + i)->is_null)
			continue;
		if((tab + i)->copied)
			GenPolyF2_free(&((tab + i)->poly));
	}
}

static int make_conv_tab(ConvPoly * tab, int size, PolyType * type, ...)
{
	int i;
	va_list ap;
	GenPolyF2 * get_poly;
	GenPolyF2 tmp;
	
	va_start(ap, type);
	for(i = 0; i < size; ++i)
	{
		if((get_poly = (GenPolyF2 *)va_arg(ap, const GenPolyF2 *)) == NULL)
		{
			(tab + i)->is_null = 1;
			continue;
		}
		(tab + i)->is_null = 0;
		(tab + i)->poly = *get_poly;
		(tab + i)->copied = 0;
		*type = (i == 0) ? (get_poly->type) : larger_type(*type, get_poly->type);
	}
	va_end(ap);
	for(i = 0; i < size; ++i)
	{
		if((tab + i)->is_null)
			continue;
		if((tab + i)->poly.type == *type)
			continue;
		if(! GenPolyF2_convert(&tmp, &((tab + i)->poly), *type))
		{
			free_conv_tab(tab, size);
			return 0;
		}
		(tab + i)->poly = tmp;
		(tab + i)->copied = 1;
	}
	return 1;
}

static int GenPolyF2_init_type(GenPolyF2 * poly, PolyType type)
{
	switch(type)
	{
		case PT_NORMAL:
		poly->create = (int (*)(void *, int))&PolyF2_create;
		poly->create_by_vals = (int (*)(void *, va_list))&PolyF2_create_by_vals_va;
		poly->create_single = (int (*)(void *, int))&PolyF2_create_single;
		poly->free = (void (*)(void *))&PolyF2_free;
		poly->set_val = (int (*)(void *, unsigned char, int))&PolyF2_set_val;
		poly->set_vals = (int (*)(void *, va_list))&PolyF2_set_vals_va;
		poly->get_val = (int (*)(const void *, unsigned char *, int))&PolyF2_get_val;
		poly->add_val = (int (*)(void *, unsigned char, int))&PolyF2_add_val;
		poly->copy = (int (*)(void *, const void *))&PolyF2_copy;
		poly->mul_degree = (int (*)(void *, int))&PolyF2_mul_degree;
		poly->add = (int (*)(void *, const void *, const void *))&PolyF2_add;
		poly->sub = (int (*)(void *, const void *, const void *))&PolyF2_sub;
		poly->mul = (int (*)(void *, const void *, const void *, const void *))&PolyF2_mul;
		poly->euclid_div = (int (*)(void *, void *, const void *, const void *))&PolyF2_euclid_div;
		poly->equals = (int (*)(const void *, const void *))&PolyF2_equals;
		poly->degree = (int (*)(const void *))&PolyF2_degree;
		poly->real_degree = (int (*)(const void *))&PolyF2_real_degree;
		poly->print = (void (*)(const void *))&PolyF2_print;
		poly->print_bin = (void (*)(const void *))&PolyF2_print_bin;
		break;
		
		case PT_INT:
		poly->create = (int (*)(void *, int))&PolyIntF2_create;
		poly->create_by_vals = (int (*)(void *, va_list))&PolyIntF2_create_by_vals_va;
		poly->create_single = (int (*)(void *, int))&PolyIntF2_create_single;
		poly->free = (void (*)(void *))&PolyIntF2_free;
		poly->set_val = (int (*)(void *, unsigned char, int))&PolyIntF2_set_val;
		poly->set_vals = (int (*)(void *, va_list))&PolyIntF2_set_vals_va;
		poly->get_val = (int (*)(const void *, unsigned char *, int))&PolyIntF2_get_val;
		poly->add_val = (int (*)(void *, unsigned char, int))&PolyIntF2_add_val;
		poly->copy = (int (*)(void *, const void *))&PolyIntF2_copy;
		poly->mul_degree = (int (*)(void *, int))&PolyIntF2_mul_degree;
		poly->add = (int (*)(void *, const void *, const void *))&PolyIntF2_add;
		poly->sub = (int (*)(void *, const void *, const void *))&PolyIntF2_sub;
		poly->mul = (int (*)(void *, const void *, const void *, const void *))&PolyIntF2_mul;
		poly->euclid_div = (int (*)(void *, void *, const void *, const void *))&PolyIntF2_euclid_div;
		poly->equals = (int (*)(const void *, const void *))&PolyIntF2_equals;
		poly->degree = (int (*)(const void *))&PolyIntF2_degree;
		poly->real_degree = (int (*)(const void *))&PolyIntF2_real_degree;
		poly->print = (void (*)(const void *))&PolyIntF2_print;
		poly->print_bin = (void (*)(const void *))&PolyIntF2_print_bin;
		break;
		
		default:
		return 0;
	}
	poly->type = type;
	return 1;
}

static void * get_poly(const GenPolyF2 * poly)
{
	switch(poly->type)
	{
		case PT_NORMAL:
		return &(((GenPolyF2 *)poly)->poly_normal);
		
		case PT_INT:
		return &(((GenPolyF2 *)poly)->poly_int);
		
		default:
		return NULL;
	}
	return NULL;
}

/* Init functions */

int GenPolyF2_create(GenPolyF2 * poly, PolyType type, int degree)
{
	GenPolyF2 res;
	if(! GenPolyF2_init_type(&res, type))
		return 0;
	if(! (*(res.create))(get_poly(&res), degree))
		return 0;
	*poly = res;
	return 1;
}

int GenPolyF2_create_by_vals(GenPolyF2 * poly, PolyType type, ...)
{
	GenPolyF2 res;
	va_list ap;
	if(! GenPolyF2_init_type(&res, type))
		return 0;
	va_start(ap, type);
	if(! (*(res.create_by_vals))(get_poly(&res), ap))
	{
		va_end(ap);
		return 0;
	}
	va_end(ap);
	*poly = res;
	return 1;
}

int GenPolyF2_create_single(GenPolyF2 * poly, PolyType type, int degree)
{
	GenPolyF2 res;
	if(! GenPolyF2_init_type(&res, type))
		return 0;
	if(! (*(res.create_single))(get_poly(&res), degree))
		return 0;
	*poly = res;
	return 1;
}

void GenPolyF2_free(GenPolyF2 * poly)
{
	(*(poly->free))(get_poly(poly));
}

int GenPolyF2_set_val(GenPolyF2 * poly, unsigned char val, int id)
{
	return (*(poly->set_val))(get_poly(poly), val, id);
}

int GenPolyF2_set_vals(GenPolyF2 * poly, ...)
{
	int success;
	va_list ap;
	
	va_start(ap, poly);
	success = (*(poly->set_vals))(get_poly(poly), ap);
	va_end(ap);
	return success;
}

int GenPolyF2_get_val(const GenPolyF2 * poly, unsigned char * val, int id)
{
	return (*(poly->get_val))(get_poly(poly), val, id);
}

int GenPolyF2_add_val(GenPolyF2 * poly, unsigned char val, int id)
{
	return (*(poly->add_val))(get_poly(poly), val, id);
}

int GenPolyF2_copy(GenPolyF2 * dest, const GenPolyF2 * src)
{
	GenPolyF2 res;
	if(! GenPolyF2_init_type(&res, src->type))
		return 0;
	if(! (*(src->copy))(get_poly(&res), get_poly(src)))
		return 0;
	*dest = res;
	return 1;
}

int GenPolyF2_convert(GenPolyF2 * dest, const GenPolyF2 * src, PolyType dest_type)
{
	GenPolyF2 res;
	int i;
	int degree;
	unsigned char val;
	
	if((degree = GenPolyF2_real_degree(src)) < 0)
		return 0;
	if(! GenPolyF2_create(&res, dest_type, degree))
		return 0;
	for(i = 0; i <= degree; ++i)
	{
		if(! GenPolyF2_get_val(src, &val, i))
			goto error_zone;
		if(! GenPolyF2_set_val(&res, val, i))
			goto error_zone;
	}
	
	*dest = res;
	return 1;
	
	error_zone:;
	GenPolyF2_free(&res);
	
	return 0;
}

int GenPolyF2_real_convert(void * dest, const GenPolyF2 * src, PolyType dest_type)
{
	GenPolyF2 tmp;
	if(! GenPolyF2_convert(&tmp, src, dest_type))
		return 0;
	switch(dest_type)
	{
		case PT_NORMAL:
		*((PolyF2 *)dest) = *((PolyF2 *)get_poly(&tmp));
		break;
		
		case PT_INT:
		*((PolyIntF2 *)dest) = *((PolyIntF2 *)get_poly(&tmp));
		break;
		
		default:
		GenPolyF2_free(&tmp);
		return 0;
	}
	return 1;
}

int GenPolyF2_copy_to_generic(GenPolyF2 * dest, void * src, PolyType src_type)
{
	GenPolyF2 res;
	if(! GenPolyF2_init_type(&res, src_type))
		return 0;
	if(! (*(res.copy))(get_poly(&res), src))
		return 0;
	*dest = res;
	return 1;
}

/* Operation functions */

int GenPolyF2_mul_degree(GenPolyF2 * poly, int degree)
{
	return (*(poly->mul_degree))(get_poly(poly), degree);
}

int GenPolyF2_add(GenPolyF2 * res, const GenPolyF2 * first, const GenPolyF2 * second)
{
	ConvPoly tab[2];
	GenPolyF2 _res;
	PolyType type;
	
	if(! make_conv_tab(tab, 2, &type, first, second))
		return 0;
	if(! GenPolyF2_init_type(&_res, type))
	{
		free_conv_tab(tab, 2);
		return 0;
	}
	if(! (*(_res.add))(get_poly(&_res), get_conv_tab(tab, 0), get_conv_tab(tab, 1)))
	{
		free_conv_tab(tab, 2);
		return 0;
	}
	
	free_conv_tab(tab, 2);
	
	*res = _res;
	return 1;
}

int GenPolyF2_sub(GenPolyF2 * res, const GenPolyF2 * first, const GenPolyF2 * second)
{
	return GenPolyF2_add(res, first, second);
}

int GenPolyF2_mul(GenPolyF2 * res, const GenPolyF2 * first, const GenPolyF2 * second, const GenPolyF2 * based_poly)
{
	ConvPoly tab[3];
	GenPolyF2 _res;
	PolyType type;
	
	if(! make_conv_tab(tab, 3, &type, first, second, based_poly))
		return 0;
	if(! GenPolyF2_init_type(&_res, type))
	{
		free_conv_tab(tab, 3);
		return 0;
	}
	if(! (*(_res.mul))(get_poly(&_res), get_conv_tab(tab, 0), get_conv_tab(tab, 1), get_conv_tab(tab, 2)))
	{
		free_conv_tab(tab, 3);
		return 0;
	}
	
	free_conv_tab(tab, 3);
	
	*res = _res;
	return 1;
}

int GenPolyF2_euclid_div(GenPolyF2 * factor, GenPolyF2 * remain, const GenPolyF2 * poly, const GenPolyF2 * divisor)
{
	ConvPoly tab[2];
	GenPolyF2 _factor;
	GenPolyF2 _remain;
	PolyType type;
	
	if(! make_conv_tab(tab, 2, &type, poly, divisor))
		return 0;
	if(! GenPolyF2_init_type(&_factor, type)
	|| ! GenPolyF2_init_type(&_remain, type))
	{
		free_conv_tab(tab, 2);
		return 0;
	}
	if(! (*(_factor.euclid_div))(get_poly(&_factor), get_poly(&_remain), get_conv_tab(tab, 0), get_conv_tab(tab, 1)))
	{
		free_conv_tab(tab, 2);
		return 0;
	}
	
	free_conv_tab(tab, 2);
	
	if(factor != NULL)
		*factor = _factor;
	else
		GenPolyF2_free(&_factor);
	if(remain != NULL)
		*remain = _remain;
	else
		GenPolyF2_free(&_remain);
	return 1;
}

int GenPolyF2_equals(const GenPolyF2 * first, const GenPolyF2 * second)
{
	int i;
	unsigned char first_val;
	unsigned char second_val;
	int degree;
	int first_degree;
	int second_degree;
	
	if((first_degree = GenPolyF2_real_degree(first)) < 0
	|| (second_degree = GenPolyF2_real_degree(second)) < 0)
		return 0;
	if(first_degree != second_degree)
		return 0;
	degree = first_degree;
	for(i = 0; i <= degree; ++i)
	{
		if(! GenPolyF2_get_val(first, &first_val, i)
		|| ! GenPolyF2_get_val(second, &second_val, i))
			return 0;
		if(first_val != second_val)
			return 0;
	}
	return 1;
}

int GenPolyF2_equals_vals(const GenPolyF2 * first, ...)
{
	GenPolyF2 res;
	int is_eq;
	va_list ap;
	
	if(! GenPolyF2_init_type(&res, first->type))
		return 0;
	va_start(ap, first);
	if(! (*(res.create_by_vals))(get_poly(&res), ap))
	{
		va_end(ap);
		return 0;
	}
	va_end(ap);
	is_eq = GenPolyF2_equals(first, &res);
	GenPolyF2_free(&res);
	return is_eq;
}

int GenPolyF2_gcd_aux(GenPolyF2 * gcd, GenPolyF2 * u, GenPolyF2 * v, const GenPolyF2 * first, const GenPolyF2 * second, const GenPolyF2 * u_i1, const GenPolyF2 * v_i1, const GenPolyF2 * u_i2, const GenPolyF2 * v_i2)
{
	int first_degree, second_degree;
	int success;
	GenPolyF2 remain, factor;
	GenPolyF2 u_i, v_i;
	GenPolyF2 tmp;
	
	if((first_degree = GenPolyF2_real_degree(first)) < 0
	|| (second_degree = GenPolyF2_real_degree(second)) < 0)
		return 0;
	if(second_degree > first_degree)
		return GenPolyF2_gcd_aux(gcd, u, v, second, first, v_i1, u_i1, v_i2, u_i2);
	
	if(second_degree == 0)
	{
		unsigned char val;
		if(! GenPolyF2_get_val(second, &val, 0))
			return 0;
		if(val == 0)
		{
			if(gcd != NULL)
			{
				if(! GenPolyF2_copy(gcd, first))
					return 0;
			}
			if(u != NULL)
			{
				if(! GenPolyF2_copy(u, u_i2))
					return 0;
			}
			if(v != NULL)
			{
				if(! GenPolyF2_copy(v, v_i2))
					return 0;
			}
			return 1;
		}
	}
	
	if(! GenPolyF2_euclid_div(&factor, &remain, first, second))
		return 0;
	
	if(! GenPolyF2_mul(&tmp, &factor, u_i1, NULL))
		goto error_zone;
	if(! GenPolyF2_sub(&u_i, u_i2, &tmp))
	{
		GenPolyF2_free(&tmp);
		goto error_zone;
	}
	GenPolyF2_free(&tmp);
	
	if(! GenPolyF2_mul(&tmp, &factor, v_i1, NULL))
	{
		GenPolyF2_free(&u_i);
		goto error_zone;
	}
	if(! GenPolyF2_sub(&v_i, v_i2, &tmp))
	{
		GenPolyF2_free(&u_i);
		GenPolyF2_free(&tmp);
		goto error_zone;
	}
	GenPolyF2_free(&tmp);
	
	success = GenPolyF2_gcd_aux(gcd, u, v, second, &remain, &u_i, &v_i, u_i1, v_i1);
	
	GenPolyF2_free(&u_i);
	GenPolyF2_free(&v_i);
	GenPolyF2_free(&remain);
	GenPolyF2_free(&factor);
	return success;
	
	error_zone:;
	GenPolyF2_free(&remain);
	GenPolyF2_free(&factor);
	return 0;
}

int GenPolyF2_gcd(GenPolyF2 * gcd, GenPolyF2 * u, GenPolyF2 * v, const GenPolyF2 * first, const GenPolyF2 * second)
{
	GenPolyF2 u_0, v_0, u_1, v_1;
	int success;
	if(! GenPolyF2_create_by_vals(&u_0, PT_NORMAL, 1, -1)
	|| ! GenPolyF2_create_by_vals(&u_1, PT_NORMAL, 0, -1)
	|| ! GenPolyF2_create_by_vals(&v_0, PT_NORMAL, 0, -1)
	|| ! GenPolyF2_create_by_vals(&v_1, PT_NORMAL, 1, -1))
		return 0;
	success = GenPolyF2_gcd_aux(gcd, u, v, first, second, &u_1, &v_1, &u_0, &v_0);
	GenPolyF2_free(&u_0);
	GenPolyF2_free(&u_1);
	GenPolyF2_free(&v_0);
	GenPolyF2_free(&v_1);
	return success;
}

int GenPolyF2_invert(GenPolyF2 * invert, const GenPolyF2 * elem, const GenPolyF2 * reference)
{
	GenPolyF2 gcd, possible_invert;
	unsigned char val;
	
	if(! GenPolyF2_gcd(&gcd, &possible_invert, NULL, elem, reference))
		return 0;
	if(GenPolyF2_real_degree(&gcd) != 0)
		goto free_zone;
	if(! GenPolyF2_get_val(&gcd, &val, 0))
		goto free_zone;
	if(val != 1)
		goto free_zone;
	*invert = possible_invert;
	GenPolyF2_free(&gcd);
	
	return 1;
	
	free_zone:;
	GenPolyF2_free(&gcd);
	GenPolyF2_free(&possible_invert);
	return 0;
}

int GenPolyF2_differentiate(GenPolyF2 * res, const GenPolyF2 * poly)
{
	int i;
	unsigned char val;
	int degree;
	
	if((degree = GenPolyF2_real_degree(poly)) <= 0)
		return 0;
	if(! GenPolyF2_create(res, PT_NORMAL, degree - 1))
		return 0;
	for(i = 1; i <= degree; ++i)
	{
		if(i % 2 == 0)
			continue;
		GenPolyF2_get_val(poly, &val, i);
		if(val == 0)
			continue;
		GenPolyF2_set_val(res, val, i - 1);
	}
	return 1;
}

int GenPolyF2_power(GenPolyF2 * res, const GenPolyF2 * poly, int power, const GenPolyF2 * based_poly)
{
	GenPolyF2 exp;
	GenPolyF2 tmp;
	if(power <= 0)
		return GenPolyF2_create_by_vals(res, PT_NORMAL, 1, -1);
	if(power == 1)
	{
		if(based_poly == NULL)
			return GenPolyF2_copy(res, poly);
		return GenPolyF2_euclid_div(NULL, res, poly, based_poly);
	}
	if(! GenPolyF2_power(&exp, poly, power / 2, based_poly))
		return 0;
	if(! GenPolyF2_mul(&tmp, &exp, &exp, based_poly))
	{
		GenPolyF2_free(&exp);
		return 0;
	}
	GenPolyF2_free(&exp);
	if(power % 2 == 0)
	{
		*res = tmp;
		return 1;
	}
	if(! GenPolyF2_mul(res, &tmp, poly, based_poly))
	{
		GenPolyF2_free(&tmp);
		return 0;
	}
	GenPolyF2_free(&tmp);
	return 1;
}

int GenPolyF2_compose(GenPolyF2 * res, const GenPolyF2 * poly, const GenPolyF2 * var, const GenPolyF2 * based_poly)
{
	GenPolyF2 _res;
	GenPolyF2 tmp_power;
	GenPolyF2 tmp_add;
	
	int poly_degree;
	int i;
	unsigned char val = 0;
	
	if((poly_degree = GenPolyF2_real_degree(poly)) < 0)
		return 0;
	if(! GenPolyF2_create_by_vals(&_res, PT_NORMAL, 0, -1))
		return 0;
	for(i = 0; i <= poly_degree; ++i)
	{
		if(! GenPolyF2_get_val(poly, &val, i))
			goto error_zone;
		if(val == 0)
			continue;
		if(! GenPolyF2_power(&tmp_power, var, i, based_poly))
			goto error_zone;
		if(! GenPolyF2_add(&tmp_add, &_res, &tmp_power))
		{
			GenPolyF2_free(&tmp_power);
			goto error_zone;
		}
		GenPolyF2_free(&tmp_power);
		GenPolyF2_free(&_res);
		_res = tmp_add;
	}
	*res = _res;
	
	return 1;
	
	error_zone:;
	
	GenPolyF2_free(&_res);
	return 0;
}

int GenPolyF2_self_add(GenPolyF2 * poly, const GenPolyF2 * add)
{
	GenPolyF2 tmp;
	if(! GenPolyF2_add(&tmp, poly, add))
		return 0;
	GenPolyF2_free(poly);
	*poly = tmp;
	return 1;
}

int GenPolyF2_self_mul(GenPolyF2 * poly, const GenPolyF2 * mul, const GenPolyF2 * based_poly)
{
	GenPolyF2 tmp;
	if(! GenPolyF2_mul(&tmp, poly, mul, based_poly))
		return 0;
	GenPolyF2_free(poly);
	*poly = tmp;
	return 1;
}

int GenPolyF2_self_power(GenPolyF2 * poly, int power, const GenPolyF2 * based_poly)
{
	GenPolyF2 tmp;
	if(! GenPolyF2_power(&tmp, poly, power, based_poly))
		return 0;
	GenPolyF2_free(poly);
	*poly = tmp;
	return 1;
}

int GenPolyF2_decal(GenPolyF2 * res, const GenPolyF2 * poly, int decal)
{
	int degree;
	int i;
	unsigned char val;
	
	degree = GenPolyF2_real_degree(poly);
	if(! GenPolyF2_create(res, PT_NORMAL, decal + degree))
		return 0;
	for(i = 0; i <= degree; ++i)
	{
		if(! GenPolyF2_get_val(poly, &val, i))
			goto error_zone;
		if(val == 0)
			continue;
		if(! GenPolyF2_set_val(res, val, i + decal))
			goto error_zone;
	}
	return 1;
	
	error_zone:;
	
	GenPolyF2_free(res);
	
	return 0;
}

int GenPolyF2_crush(GenPolyF2 * res, const GenPolyF2 * poly, int decal)
{
	int degree;
	int i;
	unsigned char val;
	
	degree = GenPolyF2_real_degree(poly);
	if(! GenPolyF2_create(res, PT_NORMAL, (decal > degree) ? 0 : (degree - decal)))
		return 0;
	for(i = 0; i <= degree - decal; ++i)
	{
		if(! GenPolyF2_get_val(poly, &val, i + decal))
			goto error_zone;
		if(val == 0)
			continue;
		if(! GenPolyF2_set_val(res, val, i))
			goto error_zone;
	}
	return 1;
	
	error_zone:;
	
	GenPolyF2_free(res);
	
	return 0;
}

/* Information functions */

int GenPolyF2_degree(const GenPolyF2 * poly)
{
	return (*(poly->degree))(get_poly(poly));
}

int GenPolyF2_real_degree(const GenPolyF2 * poly)
{
	return (*(poly->real_degree))(get_poly(poly));
}

/* In/Out functions */
#ifdef DEBUG
static void GenPolyF2_print_type(const GenPolyF2 * poly)
{
	switch(poly->type)
	{
		case PT_INT:
		printf("(type_int)");
		break;
		
		case PT_NORMAL:
		printf("(type_normal)");
		break;
		
		default:
		printf("(type_null)");
		break;
	}
}
#endif

void GenPolyF2_print(const GenPolyF2 * poly)
{
	#ifdef DEBUG
	GenPolyF2_print_type(poly);
	#endif
	(*(poly->print))(get_poly(poly));
}

void GenPolyF2_print_bin(const GenPolyF2 * poly)
{
	#ifdef DEBUG
	GenPolyF2_print_type(poly);
	#endif
	(*(poly->print_bin))(get_poly(poly));
}

void GenPolyF2_print_hexa(const GenPolyF2 * poly)
{
	int degree;
	int i;
	int j;
	unsigned char val;
	unsigned char print_val;
	degree = GenPolyF2_real_degree(poly);
	for(i = (degree / 4) * 4; i >= 0; i -= 4)
	{
		print_val = 0;
		for(j = i; j < i + 4 && j <= degree; ++j)
		{
			GenPolyF2_get_val(poly, &val, j);
			if(val)
				print_val |= (1 << (j - i));
		}
		if(print_val < 10)
			putchar('0' + print_val);
		else
			putchar('A' + print_val - 10);
	}
}

int GenPolyF2_save(const GenPolyF2 * poly, FILE * flow)
{
	if(poly->type != PT_NORMAL)
	{
		GenPolyF2 tmp;
		if(! GenPolyF2_convert(&tmp, poly, PT_NORMAL))
			return 0;
		if(! GenPolyF2_save(&tmp, flow))
		{
			GenPolyF2_free(&tmp);
			return 0;
		}
		GenPolyF2_free(&tmp);
		return 1;
	}
	return fwrite(poly->poly_normal.bit_seq, sizeof(unsigned char), size_degree(GenPolyF2_real_degree(poly)), flow) == (size_t)size_degree(GenPolyF2_real_degree(poly));
}

int GenPolyF2_load(GenPolyF2 * poly, int degree, FILE * flow)
{
	if(! GenPolyF2_create(poly, PT_NORMAL, degree))
		return 0;
	if(fread(poly->poly_normal.bit_seq, sizeof(unsigned char), size_degree(degree), flow) != (size_t)size_degree(degree))
	{
		GenPolyF2_free(poly);
		return 0;
	}
	return 1;
}

int GenPolyF2_gen_random(GenPolyF2 * poly, int degree)
{
	int i;
	if(degree < 0)
		return 0;
	if(! GenPolyF2_create(poly, PT_NORMAL, degree))
		return 0;
	if(! GenPolyF2_set_val(poly, 1, 0)
	|| ! GenPolyF2_set_val(poly, 1, degree))
		goto error_zone;
	for(i = 1; i < degree; ++i)
	{
		if(! GenPolyF2_set_val(poly, rand() % 2, i))
			goto error_zone;
	}
	return 1;
	
	error_zone:;
	GenPolyF2_free(poly);
	return 0;
}

void free_gen_tab(GenPolyF2 ** tab, int len)
{
	int i;
	if(tab == NULL || *tab == NULL)
		return;
	for(i = 0; i < len; ++i)
		GenPolyF2_free(*tab + i);
	free(*tab);
	*tab = NULL;
}

void print_gen_tab(const GenPolyF2 * tab, int len)
{
	int i;
	for(i = 0; i < len; ++i)
		GenPolyF2_print_bin(tab + i);
}