#ifndef DEF_HEADER_GENERIC_POLY
#define DEF_HEADER_GENERIC_POLY

#include "poly.h"
#include "poly_int.h"

#include <stdio.h>
#include <stdarg.h>

typedef enum
{
	PT_NULL,
	PT_INT,
	PT_NORMAL
} PolyType;

typedef struct _GenPolyF2 GenPolyF2;
struct _GenPolyF2
{
	union
	{
		PolyF2 poly_normal;
		PolyIntF2 poly_int;
	};
	PolyType type;
	
	int (*create)(void *, int degree);
	int (*create_by_vals)(void *, va_list);
	int (*create_single)(void *, int degree);
	void (*free)(void *);
	
	int (*set_val)(void *, unsigned char val, int id);
	int (*set_vals)(void *, va_list);
	int (*get_val)(const void *, unsigned char * val, int id);
	int (*add_val)(void *, unsigned char val, int id);
	int (*copy)(void * dest, const void * src);
	
	int (*mul_degree)(void *, int degree);
	int (*add)(void * res, const void * first, const void * second);
	int (*sub)(void * res, const void * first, const void * second);
	int (*mul)(void * res, const void * first, const void * second, const void * based_poly);
	int (*euclid_div)(void * factor, void * remain, const void * poly, const void * divisor);
	int (*equals)(const void *, const void *);
	
	int (*degree)(const void *);
	int (*real_degree)(const void *);
	
	void (*print)(const void *);
	void (*print_bin)(const void *);
};

/* Init functions */

int GenPolyF2_create(GenPolyF2 * poly, PolyType type, int degree);

int GenPolyF2_create_by_vals(GenPolyF2 * poly, PolyType type, ...);

int GenPolyF2_create_single(GenPolyF2 * poly, PolyType type, int degree);

void GenPolyF2_free(GenPolyF2 * poly);

int GenPolyF2_set_val(GenPolyF2 * poly, unsigned char val, int id);

int GenPolyF2_set_vals(GenPolyF2 * poly, ...);

int GenPolyF2_get_val(const GenPolyF2 * poly, unsigned char * val, int id);

int GenPolyF2_add_val(GenPolyF2 * poly, unsigned char val, int id);

int GenPolyF2_copy(GenPolyF2 * dest, const GenPolyF2 * src);

int GenPolyF2_convert(GenPolyF2 * dest, const GenPolyF2 * src, PolyType dest_type);

int GenPolyF2_real_convert(void * dest, const GenPolyF2 * src, PolyType dest_type);

int GenPolyF2_copy_to_generic(GenPolyF2 * dest, void * src, PolyType src_type);

/* Operation functions */

int GenPolyF2_mul_degree(GenPolyF2 * poly, int degree);

int GenPolyF2_add(GenPolyF2 * res, const GenPolyF2 * first, const GenPolyF2 * second);

int GenPolyF2_sub(GenPolyF2 * res, const GenPolyF2 * first, const GenPolyF2 * second);

int GenPolyF2_mul(GenPolyF2 * res, const GenPolyF2 * first, const GenPolyF2 * second, const GenPolyF2 * based_poly);

int GenPolyF2_euclid_div(GenPolyF2 * factor, GenPolyF2 * remain, const GenPolyF2 * poly, const GenPolyF2 * divisor);

int GenPolyF2_equals(const GenPolyF2 * first, const GenPolyF2 * second);

int GenPolyF2_equals_vals(const GenPolyF2 * first, ...);

int GenPolyF2_gcd(GenPolyF2 * gcd, GenPolyF2 * u, GenPolyF2 * v, const GenPolyF2 * first, const GenPolyF2 * second);

int GenPolyF2_invert(GenPolyF2 * invert, const GenPolyF2 * elem, const GenPolyF2 * reference);

int GenPolyF2_differentiate(GenPolyF2 * res, const GenPolyF2 * poly);

int GenPolyF2_power(GenPolyF2 * res, const GenPolyF2 * poly, int power, const GenPolyF2 * based_poly);

int GenPolyF2_compose(GenPolyF2 * res, const GenPolyF2 * poly, const GenPolyF2 * var, const GenPolyF2 * based_poly);

int GenPolyF2_self_add(GenPolyF2 * poly, const GenPolyF2 * add);

int GenPolyF2_self_mul(GenPolyF2 * poly, const GenPolyF2 * mul, const GenPolyF2 * based_poly);

int GenPolyF2_self_power(GenPolyF2 * poly, int power, const GenPolyF2 * based_poly);

int GenPolyF2_decal(GenPolyF2 * res, const GenPolyF2 * poly, int decal);

int GenPolyF2_crush(GenPolyF2 * res, const GenPolyF2 * poly, int decal);

/* Information functions */

int GenPolyF2_degree(const GenPolyF2 * poly);

int GenPolyF2_real_degree(const GenPolyF2 * poly);

/* In/Out functions */

void GenPolyF2_print(const GenPolyF2 * poly);

void GenPolyF2_print_bin(const GenPolyF2 * poly);

void GenPolyF2_print_hexa(const GenPolyF2 * poly);

int GenPolyF2_save(const GenPolyF2 * poly, FILE * flow);

int GenPolyF2_load(GenPolyF2 * poly, int degree, FILE * flow);

int GenPolyF2_gen_random(GenPolyF2 * poly, int degree);

/* Tabs */

void free_gen_tab(GenPolyF2 ** tab, int len);

void print_gen_tab(const GenPolyF2 * tab, int len);

#endif