#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "poly.h"
#include "color_print.h"
#include "test.h"
#include "find_primes.h"
#include "parse_poly.h"
#include "test_control.h"
#include "matrice.h"
#include "bch.h"
#include "reed-solomon.h"

#ifdef USE_QRCODE
#include "qr_code.h"
#include "qr_data.h"
#endif

void print_help()
{
	#ifndef USE_QRCODE
	printf("[--test|-t] Tests unitaires sur les fonctions de base\n");
	printf("[--test-bch|-t-bch] [POLYNOME GENERATEUR] Simulation de BCH et d'erreurs\n");
	printf("[--help|-h] Cette page\n");
	printf("[--calculator|-c] [POLYNOME GENERATEUR] Petite calculatrice:\n");
	printf("\tDe préférence parenthèsez les expressions\n");
	printf("\tOpérations binaires:\n");
	printf("\t\t+ : Addition\n");
	printf("\t\t- : Soustraction\n");
	printf("\t\t* : Multiplication (Ajouter entre [] l'expression du polynôme de référence pour modulo\n");
	printf("\t\t/ : Division\n");
	printf("\t\t%% : Modulo\n");
	printf("\t\t^ : PGCD\n");
	printf("\t\ti : inverse\n");
	printf("\t\t< : composition\n");
	printf("\tOpérations unaires:\n");
	printf("\t\t? : Test d'irreductibilité (placé avant l'expression)\n");
	printf("\t\tp : Test de primitivité (placé avant l'expression)\n");
	printf("[--simple-generate|-s] [DEGRE] : Crible d'eratosthene jusqu'à 16 par défaut et jusqu'à DEGRE si spécifié\n");
	printf("[--generate|-g] [DEGRE_MIN] [DEGRE_MAX] : Lance des tests sur des polynômes générés aléatoirement entre DEGRE_MIN et DEGRE_MAX (ou de DEGRE_MIN si DEGRE_MAX non spécifié)\n");
	printf("[--random|-r] [DEGRE] : Donne un polynôme irreductible aléatoire si déjà calculé\n");
	printf("[--list|-l] [DEGRE] : Liste tous les polynômes irreductibles d'un certain degré trouvés par la machine pour le moment\n");
	printf("[--berlekamp|-b] [POLYNOME] : Donne la décomposition d'un polynôme en les polynômes irreductibles dont il est le produit\n");
	printf("[--bch|-bch] [NB_CORRECTIONS] [POLYNOME GENERATEUR] [--code|-c|--decode|-d] [MESSAGE]\n");
	#else
	printf("[--open|-o|--load|-l] [PATH]\n");
	printf("[VERSION] [FORMAT : CORRECTION + MASK] [TEXT DATA] -s [PATH] [ZOOM]\n");
	#endif
}

int main(int argc, char * argv[])
{
	/*GenPolyF2 poly;*/
	
	#ifdef USE_QRCODE
	
	QRCode code;
	
	if(! QRCode_init())
	{
		fprintf(stderr, "Erreur lancement SDL et QRCode\n");
		exit(EXIT_FAILURE);
	}
	
	#endif
	
	srand(time(NULL));
	
	if(argc > 1)
	{
		#ifdef USE_QRCODE
		if(strcmp(argv[1], "--open") == 0 || strcmp(argv[1], "-o") == 0
		|| strcmp(argv[1], "--load") == 0 || strcmp(argv[1], "-l") == 0)
		{
			unsigned char * data = NULL;
			unsigned char * message = NULL;
			if(argc <= 2)
			{
				fprintf(stderr, "Donnez le chemin d'un fichier à ouvrir\n");
				QRCode_quit();
				exit(EXIT_FAILURE);
			}
			if(! QRCode_open(&code, argv[2]))
			{
				fprintf(stderr, "Erreur ouverture du QRCode \"%s\"\n", argv[2]);
				QRCode_quit();
				exit(EXIT_FAILURE);
			}
			QRCode_print_autozoom(&code);
			QRCode_wait_clic();
			QRCode_stop_print();
			if(! QRCode_get_format(&code))
			{
				fprintf(stderr, "Erreur de récupération du format\n");
				QRCode_quit();
				exit(EXIT_FAILURE);
			}
			if(! QRCode_decode_data(&code, &data))
			{
				fprintf(stderr, "Erreur décodage des données\n");
				QRCode_quit();
				exit(EXIT_FAILURE);
			}
			if(! get_message(data, &message))
			{
				fprintf(stderr, "Erreur récupération des données en texte\n");
				free(data);
				QRCode_quit();
				exit(EXIT_FAILURE);
			}
			free(data);
			printf("message : %s\n", message);
			free(message);
			QRCode_print_autozoom(&code);
			QRCode_wait_clic();
			QRCode_stop_print();
			QRCode_free(&code);
			
			QRCode_quit();
			exit(EXIT_SUCCESS);
		}
		/*if(strcmp(argv[1], "--create") == 0 || strcmp(argv[1], "-c") == 0)
		{
			int version;
			CorrectionLevel cl;
			int mask_id;
			if(argc <= 2)
			{
				fprintf(stderr, "Pas de version indiquée\n");
				QRCode_quit();
				exit(EXIT_FAILURE);
			}
			if(argc <= 3)
			{
				fprintf(stderr, "Pas de format indiquée\n");
				QRCode_quit();
				exit(EXIT_FAILURE);
			}
			if((cl = car_to_cl(argv[3][0])) < 0)
			{
				fprintf(stderr, "Niveau de correction %c non reconnu\n", argv[3][0]);
				QRCode_quit();
				exit(EXIT_FAILURE);
			}
			mask_id = atoi(argv[3] + 1);
			version = atoi(argv[2]);
			if(! QRCode_create_version(&code, version))
			{
				fprintf(stderr, "Erreur création du QRCode version %d\n", version);
				QRCode_quit();
				exit(EXIT_FAILURE);
			}
			if(! QRCode_set_format(&code, cl, mask_id)
			|| ! QRCode_print_format(&code))
			{
				fprintf(stderr, "Erreur écriture du format\n");
				QRCode_free(&code);
				QRCode_quit();
				exit(EXIT_FAILURE);
			}
			if(argc > 4)
			{
				char * chars = NULL;
				printf("hexa to chars %d\n", hexa_str_to_chars(argv[4], &chars));
				printf("code data %d\n", QRCode_code_data(&code, (unsigned char *)chars));
				free(chars);
			}
			QRCode_apply_mask(&code);
			QRCode_print_format(&code);
			QRCode_print_autozoom(&code);
			QRCode_wait_clic();
			QRCode_stop_print();
			QRCode_free(&code);
		}*/	
		#endif
		if(strcmp(argv[1], "--help") == 0 || strcmp(argv[1], "-h") == 0)
		{
			print_help();
			exit(EXIT_SUCCESS);
		}
		#ifndef USE_QRCODE
		if(strcmp(argv[1], "--test") == 0 || strcmp(argv[1], "-t") == 0)
			exit((full_test()) ? (EXIT_SUCCESS) : (EXIT_FAILURE));
		if(strcmp(argv[1], "--test-bch") == 0 || strcmp(argv[1], "-t-bch") == 0)
		{
			int success;
			GenPolyF2 user_poly;
			if(argc <= 2)
				exit((test_BCH()) ? (EXIT_SUCCESS) : (EXIT_FAILURE));
			if(! GenPolyF2_from_bin_str(&user_poly, argv[2], NULL))
			{
				printf("Erreur lecture du polynôme entré\n");
				exit(EXIT_FAILURE);
			}
			success = gen_test_BCH(&user_poly);
			GenPolyF2_free(&user_poly);
			exit((success) ? (EXIT_SUCCESS) : (EXIT_FAILURE));
		}
		if(strcmp(argv[1], "--calculator") == 0 || strcmp(argv[1], "-c") == 0)
		{
			if(argc <= 2)
				Calculator();
			else
			{
				GenPolyF2 based_poly;
				if(! GenPolyF2_from_bin_str(&based_poly, argv[2], NULL))
				{
					printf("Erreur lecture du polynôme entré\n");
					exit(EXIT_FAILURE);
				}
				CPoly_Calculator(&based_poly);
				GenPolyF2_free(&based_poly);
			}
			exit(EXIT_SUCCESS);
		}
		if(strcmp(argv[1], "--simple-generate") == 0 || strcmp(argv[1], "-s") == 0)
		{
			PrimePolyList list;
	
			PrimePolyList_create(&list);
			FindPolyDegree(&list, (argc > 2) ? atoi(argv[2]) : 16);
			
			/*PrimePolyList_print(&list);*/
			/*PrimePolyList_print_bin(&list);*/
			PrimePolyList_print_count(&list);
	
			PrimePolyList_free(&list);
			exit(EXIT_SUCCESS);
		}
		if(strcmp(argv[1], "--generate") == 0 || strcmp(argv[1], "-g") == 0)
		{
			int degree_min;
			int degree_max;
			if(argc <= 2)
			{
				printf("Entrez l'intervalle de recherche:\n");
				printf("%s [--generate|-g] [DEGREE_MIN] [DEGREE_MAX]\n", argv[0]);
				exit(EXIT_FAILURE);
			}
			if(argc > 2)
				degree_max = degree_min = atoi(argv[2]);
			if(argc > 3)
				degree_max = atoi(argv[3]);
			printf("Recherche de polynomes irreductibles entre %d et %d\n", degree_min, degree_max);
			exit((PrimeGeneratorBerlekamp(degree_min, degree_max)) ? EXIT_SUCCESS : EXIT_FAILURE);
		}
		if(strcmp(argv[1], "--random") == 0 || strcmp(argv[1], "-r") == 0)
		{
			GenPolyF2 poly;
			if(argc <= 2)
			{
				printf("Entrez le degré du polynôme voulu\n");
				exit(EXIT_FAILURE);
			}
			if(! GenPolyF2_load_random(&poly, atoi(argv[2])))
			{
				printf("Pas de polynôme trouvé pour le degré %d\n", atoi(argv[2]));
				exit(EXIT_FAILURE);
			}
			GenPolyF2_print_bin(&poly);
			GenPolyF2_free(&poly);
			exit(EXIT_SUCCESS);
		}
		if(strcmp(argv[1], "--list") == 0 || strcmp(argv[1], "-l") == 0)
		{
			GenPolyF2 poly;
			FILE * lec = NULL;
			if(argc <= 2)
			{
				printf("Entrez le degré des polynômes à lister\n");
				exit(EXIT_FAILURE);
			}
			if((lec = open_poly_file(atoi(argv[2]), "r")) == NULL)
			{
				printf("Liste des polynomes de degré %d non lue\n", atoi(argv[2]));
				exit(EXIT_FAILURE);
			}
			while(GenPolyF2_load(&poly, atoi(argv[2]), lec))
			{
				GenPolyF2_print_bin(&poly);
				GenPolyF2_free(&poly);
			}
			fclose(lec);
			exit(EXIT_SUCCESS);
		}
		if(strcmp(argv[1], "--berlekamp") == 0 || strcmp(argv[1], "-b") == 0)
		{
			GenPolyF2 poly;
			GenPolyF2 * factors = NULL;
			int nb_factors = 0;
			int i;
			
			if(argc <= 2)
			{
				printf("Entrez le polynôme à factoriser\n");
				exit(EXIT_FAILURE);
			}
			if(! GenPolyF2_from_bin_str(&poly, argv[2], NULL))
			{
				printf("Erreur lecture du polynôme entré\n");
				exit(EXIT_FAILURE);
			}
			if(! Berlekamp(&poly, &factors, &nb_factors))
			{
				fprintf(stderr, "Erreur lancement Berlekamp\n");
				exit(EXIT_FAILURE);
			}
			for(i = 0; i < nb_factors; ++i)
			{
				printf("(%d) : ", GenPolyF2_real_degree(factors + i));
				GenPolyF2_print_bin(factors + i);
				GenPolyF2_free(factors + i);
			}
			if(factors != NULL)
				free(factors);
			GenPolyF2_free(&poly);
			exit(EXIT_SUCCESS);
		}
		if(strcmp(argv[1], "--bch") == 0 || strcmp(argv[1], "-bch") == 0)
		{
			BCH code;
			GenPolyF2 base;
			int correct;
			if(argc <= 3)
			{
				printf("Entrez le nombre corrections voulues et le polynôme générateur du corps fini adapté\n");
				exit(EXIT_FAILURE);
			}
			correct = atoi(argv[2]);
			if(! GenPolyF2_from_bin_str(&base, argv[3], NULL))
			{
				printf("Erreur lecture du polynôme entré\n");
				exit(EXIT_FAILURE);
			}
			if(! BCH_create(&code, &base, correct))
			{
				GenPolyF2_free(&base);
				printf("Erreur création du code BCH\n");
				exit(EXIT_FAILURE);
			}
			BCH_print(&code);
			if(argc > 4)
			{
				if(strcmp(argv[4], "--code") == 0 || strcmp(argv[4], "-c") == 0)
				{
					GenPolyF2 message;
					GenPolyF2 coded_msg;
					if(argc <= 5)
					{
						printf("Message attendu en option\n");
						BCH_free(&code);
						exit(EXIT_FAILURE);
					}
					if(! GenPolyF2_from_bin_str(&message, argv[5], NULL))
					{
						printf("Erreur lecture du message\n");
						BCH_free(&code);
						exit(EXIT_FAILURE);
					}
					if(! BCH_code(&code, &coded_msg, &message))
					{
						printf("Erreur codage du message\n");
						GenPolyF2_free(&message);
						BCH_free(&code);
						exit(EXIT_FAILURE);
					}
					GenPolyF2_free(&message);
					
					printf("Message codé : ");
					GenPolyF2_print_bin(&coded_msg);
					
					GenPolyF2_free(&coded_msg);
				}
				if(strcmp(argv[4], "--decode") == 0 || strcmp(argv[4], "-d") == 0)
				{
					GenPolyF2 message;
					GenPolyF2 decoded_msg;
					if(argc <= 5)
					{
						printf("Message attendu en option\n");
						BCH_free(&code);
						exit(EXIT_FAILURE);
					}
					if(! GenPolyF2_from_bin_str(&message, argv[5], NULL))
					{
						printf("Erreur lecture du message\n");
						BCH_free(&code);
						exit(EXIT_FAILURE);
					}
					if(! BCH_decode(&code, &decoded_msg, &message))
					{
						printf("Erreur decodage du message\n");
						GenPolyF2_free(&message);
						BCH_free(&code);
						exit(EXIT_FAILURE);
					}
					GenPolyF2_free(&message);
					
					printf("Message décodé : ");
					GenPolyF2_print_bin(&decoded_msg);
					
					GenPolyF2_free(&decoded_msg);
				}
			}
			BCH_free(&code);
			exit(EXIT_SUCCESS);
		}
		/*if(strcmp(argv[1], "--reed-solomon") == 0 || strcmp(argv[1], "-rs") == 0)
		{
			RS code;
			GenPolyF2 base;
			int correct;
			
			if(argc <= 3)
			{
				printf("Entrez le nombre corrections voulues et le polynôme générateur du corps fini adapté\n");
				exit(EXIT_FAILURE);
			}
			correct = atoi(argv[2]);
			if(! GenPolyF2_from_bin_str(&base, argv[3], NULL))
			{
				printf("Erreur lecture du polynôme entré\n");
				exit(EXIT_FAILURE);
			}
			if(! RS_create(&code, &base, correct))
			{
				GenPolyF2_free(&base);
				printf("Erreur création du code BCH\n");
				exit(EXIT_FAILURE);
			}
			RS_print(&code);
			if(argc > 4)
			{
				if(strcmp(argv[4], "--code") == 0 || strcmp(argv[4], "-c") == 0)
				{
					char * chaine = NULL;
					CPoly coded_message;
					if(argc <= 5)
					{
						printf("Taille du code attendu en option\n");
						RS_free(&code);
						exit(EXIT_FAILURE);
					}
					if(argc <= 6)
					{
						printf("Message attendu en option\n");
						RS_free(&code);
						exit(EXIT_FAILURE);
					}
					printf("hsc %d\n", hexa_str_to_chars(argv[6], &chaine));
					printf("code %d\n", RS_code_chars(&code, atoi(argv[5]), (unsigned char *)chaine, &coded_message));
					free(chaine);
					CPoly_print_hexa(&coded_message);
					CPoly_print_calcul(&coded_message);
					CPoly_free(&coded_message);
				}
			}
			RS_free(&code);
			exit(EXIT_SUCCESS);
		}*/
		#endif
	}
	#ifdef USE_QRCODE
	
	QRData data_code;
	int version;
	CorrectionLevel cl;
	int mask_id;
	if(argc <= 1)
	{
		fprintf(stderr, "Pas de version indiquée\n");
		exit(EXIT_FAILURE);
	}
	if(argc <= 2)
	{
		fprintf(stderr, "Pas de format indiquée\n");
		exit(EXIT_FAILURE);
	}
	if((cl = car_to_cl(argv[2][0])) < 0)
	{
		fprintf(stderr, "Niveau de correction %c non reconnu\n", argv[2][0]);
		exit(EXIT_FAILURE);
	}
	mask_id = atoi(argv[2] + 1);
	version = atoi(argv[1]);
	if(argc <= 3)
	{
		fprintf(stderr, "Pas de message à coder\n");
		exit(EXIT_FAILURE);
	}
	if(! QRData_create_text(&data_code, argv[3]))
	{
		fprintf(stderr, "Erreur création des données du QR code\n");
		exit(EXIT_FAILURE);
	}
	QRData_print(&data_code);
	if(! QRData_make_code(&data_code, &code, version, cl, mask_id))
	{
		fprintf(stderr, "Erreur génération du QR Code\n");
		QRData_free(&data_code);
		exit(EXIT_FAILURE);
	}
	if(argc > 4 && (strcmp(argv[4], "-s") == 0 || strcmp(argv[4], "--save") == 0))
	{
		char * path = "qr_code.bmp";
		if(argc > 5)
			path = argv[5];
		if(argc > 6)
		{
			if(! QRCode_save_zoom(&code, atoi(argv[6]), path))
			{
				fprintf(stderr, "Erreur sauvegarde QR code\n");
				exit(EXIT_FAILURE);
			}
		}
		else
		{
			if(! QRCode_save(&code, path))
			{
				fprintf(stderr, "Erreur sauvegarde QR code\n");
				exit(EXIT_FAILURE);
			}
		}
	}
	QRCode_print_autozoom(&code);
	QRCode_wait_clic();
	QRCode_stop_print();
	QRCode_free(&code);
	QRData_free(&data_code);
	QRCode_quit();
	exit(EXIT_SUCCESS);
	
	#else
	
	print_help();
	
	exit(EXIT_SUCCESS);
	
	#endif
}