#include "matrice.h"

#include <stdio.h>
#include <stdlib.h>

int MatriceF2_create(MatriceF2 * mat, int lignes, int colonnes)
{
	unsigned long size = 0;
	mat->lignes = 0;
	mat->colonnes = 0;
	size = (unsigned long)lignes * (unsigned long)colonnes;
	size = (size / 8) + (size % 8 != 0);
	if((mat->seq = (unsigned char *)calloc(size, sizeof(unsigned char))) == NULL)
		return 0;
	mat->lignes = lignes;
	mat->colonnes = colonnes;
	return 1;
}

int MatriceF2_create_identity(MatriceF2 * mat, int lignes)
{
	int i;
	if(! MatriceF2_create(mat, lignes, lignes))
		return 0;
	for(i = 0; i < lignes; ++i)
		MatriceF2_set_val(mat, 1, i, i);
	return 1;
}

int MatriceF2_copy(MatriceF2 * dest, const MatriceF2 * src)
{
	unsigned long size = 0;
	unsigned long i;
	size = (unsigned long)(src->lignes) * (unsigned long)(src->colonnes);
	size = (size / 8) + (size % 8 != 0);
	if(! MatriceF2_create(dest, src->lignes, src->colonnes))
		return 0;
	for(i = 0; i < size; ++i)
		*(dest->seq + i) = *(src->seq + i);
	return 1;
}

void MatriceF2_free(MatriceF2 * mat)
{
	if(mat->seq != NULL)
		free(mat->seq);
}

unsigned long get_mat_id(const MatriceF2 * mat, int ligne, int colonne)
{
	int id;
	id = ((unsigned long)colonne * (unsigned long)(mat->lignes)) + (unsigned long)ligne;
	return id;
}

int MatriceF2_set_val(MatriceF2 * mat, unsigned char val, int ligne, int colonne)
{
	unsigned long id;
	if(ligne < 0 || ligne >= mat->lignes
	|| colonne < 0 || colonne >= mat->colonnes)
		return 0;
	id = get_mat_id(mat, ligne, colonne);
	*(mat->seq + (id / 8)) &= ~((unsigned char)1 << (id % 8));
	*(mat->seq + (id / 8)) |= (val << (id % 8));
	return 1;
}

int MatriceF2_add_val(MatriceF2 * mat, unsigned char val, int ligne, int colonne)
{
	unsigned long id;
	if(ligne < 0 || ligne >= mat->lignes
	|| colonne < 0 || colonne >= mat->colonnes)
		return 0;
	id = get_mat_id(mat, ligne, colonne);
	*(mat->seq + (id / 8)) ^= (val << (id % 8));
	return 1;
}

int MatriceF2_get_val(const MatriceF2 * mat, unsigned char * val, int ligne, int colonne)
{
	unsigned long id;
	if(ligne < 0 || ligne >= mat->lignes
	|| colonne < 0 || colonne >= mat->colonnes)
		return 0;
	id = get_mat_id(mat, ligne, colonne);
	if(val != NULL)
		*val = (*(mat->seq + (id / 8)) & (1 << (id % 8))) != 0;
	return 1;
}

int MatriceF2_poly_to_colonne(MatriceF2 * mat, const GenPolyF2 * poly, int colonne)
{
	int i;
	unsigned char val;
	for(i = 0; i < mat->lignes; ++i)
	{
		if(! GenPolyF2_get_val(poly, &val, i))
			val = 0;
		if(! MatriceF2_set_val(mat, val, i, colonne))
			return 0;
	}
	return 1;
}

int MatriceF2_poly_to_ligne(MatriceF2 * mat, const GenPolyF2 * poly, int line)
{
	int i;
	unsigned char val;
	for(i = 0; i < mat->colonnes; ++i)
	{
		if(! GenPolyF2_get_val(poly, &val, mat->lignes - i - 1))
			val = 0;
		if(! MatriceF2_set_val(mat, val, line, i))
			return 0;
	}
	return 1;
}

int MatriceF2_poly_from_colonne(const MatriceF2 * mat, GenPolyF2 * poly, int colonne)
{
	int i;
	unsigned char val;
	if(! GenPolyF2_create(poly, PT_NORMAL, mat->lignes - 1))
		return 0;
	for(i = 0; i < mat->lignes; ++i)
	{
		if(! MatriceF2_get_val(mat, &val, i, colonne))
			return 0;
		if(! GenPolyF2_set_val(poly, val, i))
			return 0;
	}
	return 1;
}

int MatriceF2_poly_from_ligne(const MatriceF2 * mat, GenPolyF2 * poly, int line)
{
	int i;
	unsigned char val;
	if(! GenPolyF2_create(poly, PT_NORMAL, mat->colonnes - 1))
		return 0;
	for(i = 0; i < mat->colonnes; ++i)
	{
		if(! MatriceF2_get_val(mat, &val, line, mat->lignes - i - 1))
			return 0;
		if(! GenPolyF2_set_val(poly, val, i))
			return 0;
	}
	return 1;
}

int poly_deg_cmp(const void * a, const void * b, void * tab)
{
	GenPolyF2 * un = (GenPolyF2 *)tab + *(int *)a;
	GenPolyF2 * deux = (GenPolyF2 *)tab + *(int *)b;
	int un_deg, deux_deg;
	un_deg = GenPolyF2_real_degree(un);
	deux_deg = GenPolyF2_real_degree(deux);
	if(un_deg != deux_deg)
		return deux_deg - un_deg;
	if(un_deg == 0)
	{
		unsigned char val_un, val_deux;
		if(! GenPolyF2_get_val(un, &val_un, 0)
		|| ! GenPolyF2_get_val(deux, &val_deux, 0))
			return 0;
		return val_deux - val_un;
	}
	return 0;
}

static int pivot_gauss(MatriceF2 * mat, MatriceF2 * calc, int data, int (*poly_from_data)(const MatriceF2 *, GenPolyF2 *, int), int (*poly_to_data)(MatriceF2 *, const GenPolyF2 *, int), int (*calc_sub)(MatriceF2 *, int, int))
{
	int i;
	int first_eq;
	int nb_moves;
	GenPolyF2 * poly_tab = NULL;
	int * ids = NULL;
	GenPolyF2 tmp;
	
	if((poly_tab = (GenPolyF2 *)malloc((data) * sizeof(GenPolyF2))) == NULL)
		return 0;
	if((ids = (int *)malloc((data) * sizeof(int))) == NULL)
	{
		free(poly_tab);
		return 0;
	}
	for(i = 0; i < data; ++i)
		*(ids + i) = i;
	do
	{
		for(i = 0; i < data; ++i)
		{
			if(! (*poly_from_data)(mat, poly_tab + i, i))
			{
				fprintf(stderr, "Polynom from vector error %d\n", i);
				goto error_zone;
			}
		}
		qsort_r(ids, data, sizeof(int), &poly_deg_cmp, poly_tab);
		nb_moves = 0;
		for(i = 1, first_eq = 0; i < data; ++i)
		{
			if(poly_deg_cmp(ids + i, ids + first_eq, poly_tab) == 0 && ! GenPolyF2_equals_vals(poly_tab + ids[i], 0, -1))
			{
				if(! GenPolyF2_sub(&tmp, poly_tab + ids[i], poly_tab + ids[first_eq]))
				{
					fprintf(stderr, "Substraction error %d\n", i);
					i = data;
					goto error_zone;
				}
				GenPolyF2_free(poly_tab + ids[i]);
				*(poly_tab + ids[i]) = tmp;
				++nb_moves;
				if(calc != NULL)
					(*calc_sub)(calc, ids[i], ids[first_eq]);
			}
			else
				first_eq = i;
		}
		for(i = 0; i < data; ++i)
		{
			if(! (*poly_to_data)(mat, poly_tab + i, i))
			{
				fprintf(stderr, "Polynom to vector error %d\n", i);
				for(; i < data; ++i)
					GenPolyF2_free(poly_tab + i);
				i = 0;
				goto error_zone;
			}
			GenPolyF2_free(poly_tab + i);
		}
	} while(nb_moves != 0);
	free(poly_tab);
	free(ids);
	return 1;
	
	error_zone:;
	for(--i; i >= 0; --i)
		GenPolyF2_free(poly_tab + i);
	free(poly_tab);
	free(ids);
	return 0;
}

int MatriceF2_apply_gauss_ligne(MatriceF2 * mat, MatriceF2 * calc)
{
	return pivot_gauss(mat, calc, mat->lignes, &MatriceF2_poly_from_ligne, &MatriceF2_poly_to_ligne, &MatriceF2_sub_lignes);
}

int MatriceF2_apply_gauss_colonne(MatriceF2 * mat, MatriceF2 * calc)
{
	return pivot_gauss(mat, calc, mat->colonnes, &MatriceF2_poly_from_colonne, &MatriceF2_poly_to_colonne, &MatriceF2_sub_colonnes);
}

int MatriceF2_sub_lignes(MatriceF2 * mat, int l1, int l2)
{
	int i;
	unsigned char val1, val2;
	for(i = 0; i < mat->colonnes; ++i)
	{
		if(! MatriceF2_get_val(mat, &val1, l1, i)
		|| ! MatriceF2_get_val(mat, &val2, l2, i))
			return 0;
		MatriceF2_set_val(mat, val1 ^ val2, l1, i);
	}
	return 1;
}

int MatriceF2_sub_colonnes(MatriceF2 * mat, int c1, int c2)
{
	int i;
	unsigned char val1, val2;
	for(i = 0; i < mat->lignes; ++i)
	{
		if(! MatriceF2_get_val(mat, &val1, i, c1)
		|| ! MatriceF2_get_val(mat, &val2, i, c2))
			return 0;
		MatriceF2_set_val(mat, val1 ^ val2, i, c1);
	}
	return 1;
}

int MatriceF2_get_ker_id(MatriceF2 * mat, GenPolyF2 ** bases, int * nb_bases)
{
	MatriceF2 aux, m;
	GenPolyF2 tmp;
	GenPolyF2 * tab = NULL;
	int i;
	int bnb = 0;
	
	if(mat->lignes != mat->colonnes)
		return 0;
	if(! MatriceF2_copy(&m, mat))
		return 0;
	if(! MatriceF2_create_identity(&aux, mat->lignes))
	{
		MatriceF2_free(&m);
		return 0;
	}
	if(! MatriceF2_apply_gauss_colonne(&m, &aux))
		goto error_zone;
	if((tab = (GenPolyF2 *)calloc(mat->colonnes, sizeof(GenPolyF2))) == NULL)
		goto error_zone;
	for(i = 0; i < mat->colonnes; ++i)
	{
		MatriceF2_poly_from_colonne(&m, &tmp, i);
		if(! GenPolyF2_equals_vals(&tmp, 0, -1))
		{
			GenPolyF2_free(&tmp);
			continue;
		}
		GenPolyF2_free(&tmp);
		MatriceF2_poly_from_colonne(&aux, tab + bnb, i);
		++bnb;
	}
	MatriceF2_free(&m);
	MatriceF2_free(&aux);
	if(bases != NULL)
		*bases = tab;
	else
	{
		for(i = 0; i < bnb; ++i)
			GenPolyF2_free(tab + i);
		free(tab);
	}
	if(nb_bases != NULL)
		*nb_bases = bnb;
	return 1;
	
	error_zone:;
	
	MatriceF2_free(&m);
	MatriceF2_free(&aux);
	return 0;
}

int MatriceF2_Berlekamp(MatriceF2 * mat, const GenPolyF2 * poly)
{
	GenPolyF2 start;
	GenPolyF2 tmp;
	GenPolyF2 xp;
	int degree;
	int i;
	
	if((degree = GenPolyF2_real_degree(poly)) <= 0)
		return 0;
	if(! GenPolyF2_create_by_vals(&start, PT_NORMAL, 1, -1))
		return 0;
	if(! GenPolyF2_create_by_vals(&xp, PT_NORMAL, 0, 0, 1, -1))
	{
		GenPolyF2_free(&start);
		return 0;
	}
	if(! MatriceF2_create(mat, degree, degree))
	{
		GenPolyF2_free(&xp);
		GenPolyF2_free(&start);
		return 0;
	}
	for(i = 1; i < degree; ++i)
	{
		if(! GenPolyF2_mul(&tmp, &start, &xp, poly))
			goto error_zone;
		GenPolyF2_free(&start);
		start = tmp;
		MatriceF2_poly_to_colonne(mat, &start, i);
		MatriceF2_add_val(mat, 1, i, i);
	}
	
	GenPolyF2_free(&xp);
	GenPolyF2_free(&start);
	
	return 1;
	
	error_zone:;
	GenPolyF2_free(&xp);
	GenPolyF2_free(&start);
	MatriceF2_free(mat);
	
	return 0;
}

void MatriceF2_print(const MatriceF2 * mat)
{
	int i, j;
	unsigned char val = 0;
	printf("MatriceF2 %d x %d\n", mat->lignes, mat->colonnes);
	for(i = 0; i < mat->lignes; ++i)
	{
		for(j = 0; j < mat->colonnes; ++j)
		{
			MatriceF2_get_val(mat, &val, i, j);
			putchar('0' + val);
		}
		putchar('\n');
	}
}

int MatriceF2_is_null_colonne(const MatriceF2 * mat)
{
	int i, j;
	unsigned char val;
	int nullc;
	
	for(i = 0; i < mat->colonnes; ++i)
	{
		nullc = 1;
		for(j = 0; j < mat->lignes; ++j)
		{
			MatriceF2_get_val(mat, &val, j, i);
			if(val != 0)
			{
				nullc = 0;
				break;
			}
		}
		if(nullc)
			return 1;
	}
	return 0;
}