#include "parse_poly.h"

#include <stdio.h>
#include <stdlib.h>
#include <readline/readline.h>
#include <readline/history.h>
#include "find_primes.h"

static int in_chaine(int car, const char * chaine)
{
	for(; *chaine != '\0'; ++chaine)
	{
		if(car == *chaine)
			return 1;
	}
	return 0;
}

static int is_sep(int car)
{
	return in_chaine(car, " \t\n");
}

static int is_binary_digit(int car)
{
	return in_chaine(car, "01");
}

static int is_number(int car)
{
	return car >= '0' && car <= '9';
}

static int is_hexa(int car)
{
	return is_number(car) || (car >= 'A' && car <= 'F') || (car >= 'a' && car <= 'f');
}

static int get_hexa(int car)
{
	if(is_number(car))
		return car - '0';
	if(car >= 'A' && car <= 'F')
		return car - 'A' + 10;
	return car - 'a' + 10;
}

static int is_bin_car(int car)
{
	if(is_sep(car))
		return 1;
	if(is_binary_digit(car))
		return 1;
	return 0;
}

/*static int is_prefixe(const char * pref, const char * chaine)
{
	for(; *pref != '\0'; ++pref, ++chaine)
	{
		if(*pref != *chaine)
			return 0;
	}
	return 1;
}*/

static const char * after_sep(const char * line)
{
	for(; is_sep(*line); ++line);
	return line;
}

int GenPolyF2_from_bin_str(GenPolyF2 * res, const char * line, const char ** end_ptr)
{
	int degree;
	const char * cur = NULL;
	GenPolyF2 poly;
	
	line = after_sep(line);
	
	if(! is_bin_car(*line))
		return GenPolyF2_from_hexa_str(res, line, end_ptr);
	
	for(cur = line, degree = 0; *cur != '\0' && is_bin_car(*cur); ++cur)
	{
		degree += is_binary_digit(*cur);
	}
	
	if(end_ptr != NULL)
		*end_ptr = cur;
	if(! GenPolyF2_create(&poly, PT_NORMAL, degree))
		return 0;
	--degree;
	for(cur = line; *cur != '\0' && is_bin_car(*cur); ++cur)
	{
		if(! is_binary_digit(*cur))
			continue;
		if(! GenPolyF2_set_val(&poly, *cur - '0', degree))
		{
			GenPolyF2_free(&poly);
			return 0;
		}
		--degree;
	}
	
	*res = poly;
	return 1;
}

int GenPolyF2_from_hexa_str(GenPolyF2 * res, const char * line, const char ** end_ptr)
{
	int degree;
	const char * cur = NULL;
	GenPolyF2 poly;
	int val;
	int i;
	
	if(in_chaine(*line, "hH"))
		++line;
	
	for(cur = line, degree = 0; *cur != '\0' && (is_hexa(*cur) || is_sep(*cur)); ++cur)
	{
		degree += is_hexa(*cur) * 4;
	}
	
	if(end_ptr != NULL)
		*end_ptr = cur;
	if(! GenPolyF2_create(&poly, PT_NORMAL, degree))
		return 0;
	degree -= 4;
	for(cur = line; *cur != '\0' && (is_hexa(*cur) || is_sep(*cur)); ++cur)
	{
		if(! is_hexa(*cur))
			continue;
		val = get_hexa(*cur);
		for(i = 0; i < 4; ++i)
		{
			if(! GenPolyF2_set_val(&poly, (val & (1 << i)) != 0, degree + i))
			{
				GenPolyF2_free(&poly);
				return 0;
			}
		}
		degree -= 4;
	}
	
	*res = poly;
	return 1;
}

static const char * end_par(const char * line)
{
	int nb = 0;
	if(*line != '(')
		return line;
	nb = 1;
	while(nb > 0)
	{
		++line;
		switch(*line)
		{
			case '(':
			++nb;
			break;
			
			case ')':
			--nb;
			break;
			
			default:
			break;
		}
	}
	return line;
}

static int sub_calcul(GenPolyF2 * res, const char * line, const char ** next)
{
	if(*line == '(')
	{
		*next = end_par(line) + 1;
		return GenPolyF2_calcul(res, line + 1, end_par(line));
	}
	return GenPolyF2_from_bin_str(res, line, next);
}

static int is_op(const char * op)
{
	switch(*op)
	{
		case '+':
		case '-':
		case '*':
		case '/':
		case '%':
		case '^':
		case '<':
		case 'i':
		return 1;
		
		default:
		break;
	}
	
	return 0;
}

static int op_ajout(const char * op)
{
	switch(*op)
	{
		case '*':
		case '<':
		return 1;
		
		default:
		break;
	}
	
	return 0;
}

static int combinable_op(const char * op)
{
	switch(*op)
	{
		case '+':
		case '-':
		case '*':
		return 1;
		
		default:
		break;
	}
	
	return 0;
}

static int make_op(GenPolyF2 * res, const char * op, GenPolyF2 * first, GenPolyF2 * second, GenPolyF2 * ajout)
{
	/*GenPolyF2_print_bin(first);
	printf("%c\n", *op);
	GenPolyF2_print_bin(second);
	if(ajout != NULL)
		GenPolyF2_print_bin(ajout);*/
		
	switch(*op)
	{
		case '+':
		return GenPolyF2_add(res, first, second);
		
		case '-':
		return GenPolyF2_sub(res, first, second);
		
		case '*':
		return GenPolyF2_mul(res, first, second, ajout);
		
		case '/':
		return GenPolyF2_euclid_div(res, NULL, first, second);
		
		case '%':
		return GenPolyF2_euclid_div(NULL, res, first, second);
		
		case '^':
		{
			int success;
			GenPolyF2 u, v;
			success = GenPolyF2_gcd(res, &u, &v, first, second);
			printf("u : ");
			GenPolyF2_print_bin(&u);
			printf("v : ");
			GenPolyF2_print_bin(&v);
			GenPolyF2_free(&u);
			GenPolyF2_free(&v);
			return success;
		}
		break;
		
		case '<':
		return GenPolyF2_compose(res, first, second, ajout);
		
		case 'i':
		return GenPolyF2_invert(res, first, second);
		
		default:
		fprintf(stderr, "'%c' opération inconnue\n", *op);
		return 0;
	}
	return 1;
}

int GenPolyF2_calcul(GenPolyF2 * res, const char * line, const char * next)
{
	GenPolyF2 first;
	GenPolyF2 second;
	GenPolyF2 ajout;
	
	const char * op = NULL;
	int success = 0;
	
	line = after_sep(line);
	
	if(*line == '?')
	{
		++line;
		line = after_sep(line);
		if(! sub_calcul(res, line, &line))
		{
			fprintf(stderr, "Erreur lecture premier argument\n");
			return 0;
		}
		printf("%s\n", (PrimeTest(res)) ? "Irreductible" : "Produit de polynomes");
		return 1;
	}
	if(*line == 'p')
	{
		++line;
		line = after_sep(line);
		if(! sub_calcul(res, line, &line))
		{
			fprintf(stderr, "Erreur lecture premier argument\n");
			return 0;
		}
		printf("%s\n", (PrimitiveTest(res)) ? "Primitif" : "Non Primitif");
		return 1;
	}
	
	if(! sub_calcul(&first, line, &line))
	{
		fprintf(stderr, "Erreur lecture premier argument\n");
		return 0;
	}
		
	line = after_sep(line);
	if(*line == '\0' || line == next)
	{
		*res = first;
		return 1;
	}
	
	if(! is_op(op = line))
	{
		fprintf(stderr, "'%c' n'est pas opérateur\n", *op);
		return 0;
	}
	++line;
	
	line = after_sep(line);
	
	if(! sub_calcul(&second, line, &line))
	{
		GenPolyF2_free(&first);
		fprintf(stderr, "Erreur lecture second argument\n");
		return 0;
	}
	
	line = after_sep(line);
	
	if(*line != '[' || line == next)
	{
		GenPolyF2 sub_res;
		GenPolyF2 sub_arg;
		success = make_op(res, op, &first, &second, NULL);
		while(*line == *op && combinable_op(op))
		{
			++line;
			line = after_sep(line);
			if(! sub_calcul(&sub_arg, line, &line))
			{
				fprintf(stderr, "Erreur lecture argument supplémentaire\n");
				return 0;
			}
			success &= make_op(&sub_res, op, res, &sub_arg, NULL);
			if(! success)
				break;
			GenPolyF2_free(res);
			GenPolyF2_free(&sub_arg);
			*res = sub_res;
		}
	}
	if(*line == '[' && line != next)
	{
		if(! op_ajout(op))
		{
			fprintf(stderr, "'%c' ne laisse pas la possibilité de données complémentaires\n", *op);
			GenPolyF2_free(&first);
			GenPolyF2_free(&second);
			return 0;
		}
		if(! sub_calcul(&ajout, line + 1, &line))
		{
			fprintf(stderr, "Erreur lecture ajout\n");
			GenPolyF2_free(&first);
			GenPolyF2_free(&second);
			return 0;
		}
		if(*line != ']')
		{
			fprintf(stderr, "L'ajout n'est pas fermé par ']' correctement\n");
			GenPolyF2_free(&first);
			GenPolyF2_free(&second);
			GenPolyF2_free(&ajout);
			return 0;
		}
		++(line);
		success = make_op(res, op, &first, &second, &ajout);
		GenPolyF2_free(&ajout);
	}
		
		
	GenPolyF2_free(&first);
	GenPolyF2_free(&second);
	
	return success;
}

int Calculator()
{
	char * line = NULL;
	GenPolyF2 poly;
	
	while((line = readline(NULL)) != NULL)
	{
		add_history(line);
		if(! GenPolyF2_calcul(&poly, line, line + strlen(line)))
		{
			fprintf(stderr, "Erreur calcul\n");
			free(line);
			continue;
		}
		printf("(%d) : ", GenPolyF2_real_degree(&poly));
		GenPolyF2_print_bin(&poly);
		GenPolyF2_free(&poly);
		free(line);
	}
	return 1;
}

/* CPoly calculator */

static int is_CPop(const char * op)
{
	switch(*op)
	{
		case '+':
		case '*':
		case '/':
		case '%':
		case '^':
		case '<':
		return 1;
		
		default:
		break;
	}
	
	return 0;
}

static int combinable_CPop(const char * op)
{
	switch(*op)
	{
		case '+':
		case '*':
		return 1;
		
		default:
		break;
	}
	
	return 0;
}

static int make_CPop(CPoly * res, const char * op, CPoly * first, CPoly * second)
{		
	switch(*op)
	{
		case '+':
		return CPoly_add(res, first, second);
		
		case '*':
		return CPoly_mul(res, first, second);
		
		case '/':
		return CPoly_euclid_div(res, NULL, first, second);
		
		case '%':
		return CPoly_euclid_div(NULL, res, first, second);
		
		case '^':
		{
			CPoly u, v;
			if(! CPoly_gcd(res, &u, &v, first, second))
				return 0;
			printf("U(X) : ");
			CPoly_print_calcul(&u);
			printf("V(X) : ");
			CPoly_print_calcul(&v);
			CPoly_free(&u);
			CPoly_free(&v);
			return 1;
		}
		
		case '<':
		return CPoly_compose(res, first, second);
		
		default:
		fprintf(stderr, "'%c' opération inconnue\n", *op);
		return 0;
	}
	return 1;
}

static int char_to_int_in(const char * start, const char * end)
{
	int val = 0;
	while(start != end && (is_number(*start) || is_sep(*start)))
	{
		if(is_number(*start))
		{
			val *= 10;
			val += (*start - '0');
		}
		++start;
	}
	return val;
}

static int get_degree(const char * line)
{
	int degree = -1;
	do
	{
		line = after_sep(line);
		while(is_bin_car(*line) || is_hexa(*line) || in_chaine(*line, "hH"))
			++line;
		line = after_sep(line);
		if(in_chaine(*line, "xX"))
		{
			const char * start = NULL;
			int tmp_deg = 1;
			++line;
			line = after_sep(line);
			start = line;
			while(is_number(*line) || is_sep(*line))
				++line;
			tmp_deg = char_to_int_in(start, line);
			if(tmp_deg == 0)
				tmp_deg = 1;
			if(tmp_deg > degree)
				degree = tmp_deg;
		}
		else if(degree < 0)
			degree = 0;
		line = after_sep(line);
		if(*line != '.')
			return degree;
		++line;
	} while(1);
	return -1;
}

int CPoly_from_str(CPoly * res, const GenPolyF2 * based_poly, const char * line, const char ** end_ptr)
{
	GenPolyF2 tmp;
	int degree;
	line = after_sep(line);
	if((degree = get_degree(line)) < 0)
		return 0;
	if(! CPoly_create(res, degree, based_poly))
		return 0;
	do
	{
		line = after_sep(line);
		if(is_bin_car(*line) || is_hexa(*line) || in_chaine(*line, "hH"))
		{
			GenPolyF2 sub_tmp;
			if(! GenPolyF2_from_bin_str(&tmp, line, &line))
			{
				CPoly_free(res);
				return 0;
			}
			if(! GenPolyF2_euclid_div(NULL, &sub_tmp, &tmp, based_poly))
			{
				GenPolyF2_free(&tmp);
				CPoly_free(res);
				return 0;
			}
			GenPolyF2_free(&tmp);
			tmp = sub_tmp;
		}
		else
		{
			if(! GenPolyF2_create_single(&tmp, PT_NORMAL, 0))
			{
				CPoly_free(res);
				return 0;
			}
		}
		line = after_sep(line);
		if(in_chaine(*line, "xX"))
		{
			const char * start = NULL;
			++line;
			line = after_sep(line);
			start = line;
			while(is_number(*line) || is_sep(*line))
				++line;
			degree = char_to_int_in(start, line);
			if(degree == 0)
				degree = 1;
		}
		else
			degree = 0;
		if(! CPoly_set_val(res, &tmp, degree))
		{
			GenPolyF2_free(&tmp);
			CPoly_free(res);
			return 0;
		}
		GenPolyF2_free(&tmp);
		line = after_sep(line);
		if(*line != '.')
			break;
		++line;
	} while(1);
	if(end_ptr != NULL)
		*end_ptr = line;
	return 1;
}

static int CPoly_get_elem(CPoly * res, const GenPolyF2 * based_poly, const char * line, const char ** next)
{
	if(*line == '(')
	{
		*next = end_par(line) + 1;
		return CPoly_calcul(res, based_poly, line + 1, end_par(line));
	}
	return CPoly_from_str(res, based_poly, line, next);
}

int CPoly_calcul(CPoly * res, const GenPolyF2 * based_poly, const char * line, const char * next)
{
	CPoly first;
	CPoly second;
	
	const char * op = NULL;
	int success = 1;
	
	line = after_sep(line);
	
	if(! CPoly_get_elem(&first, based_poly, line, &line))
	{
		fprintf(stderr, "Erreur lecture premier argument\n");
		return 0;
	}
		
	line = after_sep(line);
	if(*line == '\0' || line == next)
	{
		*res = first;
		return 1;
	}
	
	if(! is_CPop(op = line))
	{
		fprintf(stderr, "'%c' n'est pas opérateur\n", *op);
		return 0;
	}
	++line;
	
	line = after_sep(line);
	
	if(! CPoly_get_elem(&second, based_poly, line, &line))
	{
		CPoly_free(&first);
		fprintf(stderr, "Erreur lecture second argument\n");
		return 0;
	}
	
	line = after_sep(line);
	
	{
		CPoly sub_res;
		CPoly sub_arg;
		success = make_CPop(res, op, &first, &second);
		
		while(line != next && *line == *op && combinable_CPop(op))
		{
			++line;
			line = after_sep(line);
			if(! CPoly_get_elem(&sub_arg, based_poly, line, &line))
			{
				fprintf(stderr, "Erreur lecture argument supplémentaire\n");
				return 0;
			}
			success &= make_CPop(&sub_res, op, res, &sub_arg);
			if(! success)
				break;
			CPoly_free(res);
			CPoly_free(&sub_arg);
			*res = sub_res;
			line = after_sep(line);
		}
	}
	
	CPoly_free(&first);
	CPoly_free(&second);
	
	return success;
}

int CPoly_Calculator(const GenPolyF2 * based_poly)
{
	char * line = NULL;
	CPoly poly;
	
	while((line = readline(NULL)) != NULL)
	{
		add_history(line);
		if(! CPoly_calcul(&poly, based_poly, line, line + strlen(line)))
		{
			fprintf(stderr, "Erreur calcul\n");
			free(line);
			continue;
		}
		CPoly_print_calcul(&poly);
		CPoly_free(&poly);
		free(line);
	}
	return 1;
}