#ifndef DEF_HEADER_POLY
#define DEF_HEADER_POLY

#include <stdarg.h>

typedef struct _PolyF2 PolyF2;
struct _PolyF2
{
	unsigned char * bit_seq;
	int degree;
};

int size_degree(int degree);

/* Init functions */

int PolyF2_create(PolyF2 * poly, int degree);

int PolyF2_create_by_vals(PolyF2 * poly, ...);

int PolyF2_create_by_vals_va(PolyF2 * poly, va_list ap);

int PolyF2_create_single(PolyF2 * poly, int degree);

void PolyF2_free(PolyF2 * poly);

int PolyF2_set_val(PolyF2 * poly, unsigned char val, int id);

int PolyF2_set_vals(PolyF2 * poly, ...);

int PolyF2_set_vals_va(PolyF2 * poly, va_list ap);

int PolyF2_get_val(const PolyF2 * poly, unsigned char * val, int id);

int PolyF2_add_val(PolyF2 * poly, unsigned char val, int id);

int PolyF2_copy(PolyF2 * dest, const PolyF2 * src);

/* Operation functions */

int PolyF2_mul_degree(PolyF2 * poly, int degree);

int PolyF2_add(PolyF2 * res, const PolyF2 * first, const PolyF2 * second);

int PolyF2_sub(PolyF2 * res, const PolyF2 * first, const PolyF2 * second);

int PolyF2_mul(PolyF2 * res, const PolyF2 * first, const PolyF2 * second, const PolyF2 * based_poly);

int PolyF2_euclid_div(PolyF2 * factor, PolyF2 * remain, const PolyF2 * poly, const PolyF2 * divisor);

int PolyF2_equals(const PolyF2 * first, const PolyF2 * second);

/* Information functions */

int PolyF2_degree(const PolyF2 * poly);

int PolyF2_real_degree(const PolyF2 * poly);

/* In/Out functions */

void PolyF2_print(const PolyF2 * poly);

void PolyF2_print_bin(const PolyF2 * poly);

#endif