#ifndef DEF_HEADER_POLY_INT
#define DEF_HEADER_POLY_INT

#include <inttypes.h>
#include <stdarg.h>

#define BS_TYPE unsigned int

typedef struct _PolyIntF2 PolyIntF2;
struct _PolyIntF2
{
	BS_TYPE bit_seq;
	int degree;
};

/* Init functions */

int PolyIntF2_create(PolyIntF2 * poly, int degree);

int PolyIntF2_create_by_vals(PolyIntF2 * poly, ...);

int PolyIntF2_create_by_vals_va(PolyIntF2 * poly, va_list ap);

int PolyIntF2_create_single(PolyIntF2 * poly, int degree);

void PolyIntF2_free(PolyIntF2 * poly);

int PolyIntF2_set_val(PolyIntF2 * poly, unsigned char val, int id);

int PolyIntF2_set_vals(PolyIntF2 * poly, ...);

int PolyIntF2_set_vals_va(PolyIntF2 * poly, va_list ap);

int PolyIntF2_get_val(const PolyIntF2 * poly, unsigned char * val, int id);

int PolyIntF2_add_val(PolyIntF2 * poly, unsigned char val, int id);

int PolyIntF2_copy(PolyIntF2 * dest, const PolyIntF2 * src);

/* Operation functions */

int PolyIntF2_mul_degree(PolyIntF2 * poly, int degree);

int PolyIntF2_add(PolyIntF2 * res, const PolyIntF2 * first, const PolyIntF2 * second);

int PolyIntF2_sub(PolyIntF2 * res, const PolyIntF2 * first, const PolyIntF2 * second);

int PolyIntF2_mul(PolyIntF2 * res, const PolyIntF2 * first, const PolyIntF2 * second, const PolyIntF2 * based_poly);

int PolyIntF2_euclid_div(PolyIntF2 * factor, PolyIntF2 * remain, const PolyIntF2 * poly, const PolyIntF2 * divisor);

int PolyIntF2_equals(const PolyIntF2 * first, const PolyIntF2 * second);

/* Information functions */

int PolyIntF2_degree(const PolyIntF2 * poly);

int PolyIntF2_real_degree(const PolyIntF2 * poly);

/* In/Out functions */

void PolyIntF2_print(const PolyIntF2 * poly);

void PolyIntF2_print_bin(const PolyIntF2 * poly);

#endif