#ifndef DEF_HEADER_QR_CODE
#define DEF_HEADER_QR_CODE

#include <SDL/SDL.h>

#include "generic_poly.h"
#include "reed-solomon.h"

typedef enum
{
	L = 0x01,
	M = 0x00,
	Q = 0x03,
	H = 0x02
} CorrectionLevel;

typedef struct _QRCode QRCode;
struct _QRCode
{
	SDL_Surface * surface;
	SDL_Surface * area;
	
	SDL_Rect current_coord;
	int current_id;
	
	CorrectionLevel cl;
	int mask_id;
	int version;
	
	int (*get_next_coord)(SDL_Rect *, QRCode *, void * user_data);
	int (*get_coord_by_id)(int id, SDL_Rect *, QRCode *, void * user_data);
	int (*is_message_end)(QRCode *, void * user_data);
	void * user_data_move;
	
	RS rs;
	int nb_codewords;
};

/* Global init */

int QRCode_init();

void QRCode_quit();

/* Struct init */

int QRCode_create(QRCode * code, int size);

int QRCode_open(QRCode * code, const char * path);

void QRCode_free(QRCode * code);

int QRCode_set_move(QRCode * code, int (*get_next_coord)(SDL_Rect *, QRCode *, void *),
	int (*get_coord_by_id)(int id, SDL_Rect *, QRCode *, void *),
	int (*is_message_end)(QRCode *, void *), int (*init_data)(QRCode *, void *),
	void * user_data);
	
int QRCode_go_to_next(QRCode * code);

int QRCode_read_bit(QRCode * code, unsigned char * bit);

int QRCode_write_bit(QRCode * code, unsigned char bit);

int QRCode_read_data(QRCode * code, void * data, int bit_size,
	int (*set_data)(void * data, int current_id, unsigned char bit));

int QRCode_write_data(QRCode * code, void * data, int bit_size,
	int (*get_data)(void * data, int current_id, unsigned char * bit));

int QRCode_read_GenPolyF2(QRCode * code, GenPolyF2 * poly, int degree);

int QRCode_write_GenPolyF2(QRCode * code, const GenPolyF2 * poly, int degree);

int QRCode_read_CPoly(QRCode * code, CPoly * poly, int degree);

int QRCode_write_CPoly(QRCode * code, CPoly * poly, int degree);

int QRCode_print(QRCode * code);

int QRCode_print_zoom(QRCode * code, int scale);

void QRCode_print_autozoom(QRCode * code);

int QRCode_save(QRCode * code, const char * path);

int QRCode_save_zoom(QRCode * code, int scale, const char * path);

int QRCode_save_autozoom(QRCode * code, const char * path);

int QRCode_wait_clic();

void QRCode_stop_print();

CorrectionLevel car_to_cl(char c);

int get_nb_codewords(int version, CorrectionLevel cl);

int get_nb_error_codewords(int version, CorrectionLevel cl);

int get_capacity(int version, CorrectionLevel cl);

#ifdef ANIM
void print_area(QRCode * code);
#endif

/* Graphics */

int QRCode_create_version(QRCode * code, int version);

int QRCode_set_format(QRCode * code, CorrectionLevel cl, int mask_id);

int QRCode_code_data(QRCode * code, const unsigned char * data);

int QRCode_decode_data(QRCode * code, unsigned char ** data);

void QRCode_print_align(QRCode * code);

int QRCode_print_format(QRCode * code);

int QRCode_get_format(QRCode * code);

void QRCode_apply_mask(QRCode * code);

/* Image */

int Image_to_matrix(const SDL_Surface * image, SDL_Surface ** matrix);

#endif