#include "qr_data.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int QRData_create(QRData * data, QRD_Mode mode, int data_size)
{
	data->mode = mode;
	data->infos = 0;
	data->data_size = 0;
	data->final_data = NULL;
	if((data->data = (unsigned char *)calloc(data_size, sizeof(unsigned char))) == NULL)
		return 0;
	data->data_size = data_size;
	data->final_data_size = 0;
	return 1;
}

void QRData_free(QRData * data)
{
	data->data_size = 0;
	data->final_data_size = 0;
	if(data->data != NULL)
	{
		free(data->data);
		data->data = NULL;
	}
	if(data->final_data != NULL)
	{
		free(data->final_data);
		data->final_data = NULL;
	}
}

static int power_int(int a, int n)
{
	if(n <= 0)
		return 1;
	return a * power_int(a, n - 1);
}

/*static void set_bits(void * sequence, void * add, int add_size, int offset)
{
	int i;
	unsigned char * seq = (unsigned char *)sequence;
	unsigned char * add_seq = (unsigned char *)add;
	seq += (offset / 8);
	offset %= 8;
	for(i = 0; i < add_size; ++i)
	{
		*(seq + ((offset + i) / 8)) &= ~((unsigned char)1 << ((offset + i) % 8));
		*(seq + ((offset + i) / 8)) |=
			((unsigned char)((*(add_seq + (i / 8)) & (1 << (i % 8))) != 0) << ((offset + i) % 8));
	}
}*/

static void set_bits_rev(void * sequence, void * add, int add_size, int offset)
{
	int i;
	int k;
	unsigned char * seq = (unsigned char *)sequence;
	unsigned char * add_seq = (unsigned char *)add;
	seq += (offset / 8);
	offset %= 8;
	for(i = add_size - 1; i >= 0; --i)
	{
		k = 7 - ((offset + (add_size - i - 1)) % 8);
		*(seq + ((offset + add_size - i - 1) / 8)) &= ~((unsigned char)1 << k);
		*(seq + ((offset + add_size - i - 1) / 8)) |=
			((unsigned char)((*(add_seq + (i / 8)) & (1 << (i % 8))) != 0) << k);
	}
}

static void get_bits_rev(const void * sequence, void * add, int add_size, int offset)
{
	int i;
	int k;
	unsigned char * seq = (unsigned char *)sequence;
	unsigned char * add_seq = (unsigned char *)add;
	seq += (offset / 8);
	offset %= 8;
	for(i = add_size - 1; i >= 0; --i)
	{
		k = 7 - ((offset + (add_size - i - 1)) % 8);
		*(add_seq + (i / 8)) &= ~((unsigned char)1 << (i % 8));
		*(add_seq + (i / 8)) |=
			(unsigned char)(((*(seq + ((offset + add_size - i - 1) / 8)) & (1 << k)) != 0) << (i % 8));
	}
}

static int info_bits_size(QRD_Mode mode)
{
	switch(mode)
	{
		case QRD_NUMERIC:
		return 10;
		
		case QRD_ALPHANUMERIC:
		return 9;
		
		case QRD_BYTE:
		return 8;
		
		default:
		break;
	}
	return 8;
}

static int get_alphanumeric_val(int val)
{
	const char * chaine = " $%*+-./:";
	if(val >= 0 && val <= 9)
		return val + '0';
	if(val >= 10 && val <= 35)
		return val - 10 + 'A';
	if(val >= 36 && val <= 44)
		return *(chaine + (val - 36));
	return -1;
}

int get_message(const unsigned char * data, unsigned char ** message)
{
	QRD_Mode mode = 0;
	unsigned int infos = 0;
	int bit_info;
	int i, j;
	unsigned char * res = NULL;
	unsigned int tmp;
	
	get_bits_rev(data, &mode, 4, 0);
	bit_info = info_bits_size(mode);
	get_bits_rev(data, &infos, bit_info, 4);
	if((res = (unsigned char *)calloc(infos + 1, sizeof(unsigned char))) == NULL)
		return 0;
	/*printf("%d -> %d\n", mode, infos);*/
	switch(mode)
	{
		case QRD_NUMERIC:
		printf("Mode Numerique\n");
		for(i = 0; i < infos / 3; ++i)
		{
			tmp = 0;
			get_bits_rev(data, &tmp, 10, 4 + bit_info + (i * 10));
			for(j = 0; j < 3; ++j)
			{
				*(res + (i * 3) + (2 - j)) = '0' + tmp % 10;
				tmp /= 10;
			}
		}
		if(infos % 3 != 0)
		{
			tmp = 0;
			get_bits_rev(data, &tmp, 1 + (3 * (infos % 3)), 4 + bit_info + ((infos / 3) * 10));
			for(j = 0; j < (infos % 3); ++j)
			{
				*(res + ((infos / 3) * 3) + ((infos % 3) - 1 - j)) = '0' + tmp % 10;
				tmp /= 10;
			}
		}
		break;
		
		case QRD_ALPHANUMERIC:
		printf("Mode Alphanumerique\n");
		for(i = 0; i < infos / 2; ++i)
		{
			tmp = 0;
			get_bits_rev(data, &tmp, 11, 4 + bit_info + (i * 11));
			*(res + (i * 2)) = get_alphanumeric_val(tmp / 45);
			*(res + (i * 2) + 1) = get_alphanumeric_val(tmp % 45);
		}
		if(infos % 2 == 1)
		{
			tmp = 0;
			get_bits_rev(data, &tmp, 6, 4 + bit_info + ((infos / 2) * 11));
			*(res + (infos - 1)) = get_alphanumeric_val(tmp);
		}
		break;
		
		case QRD_BYTE:
		printf("Mode Byte\n");
		for(i = 0; i < infos; ++i)
			get_bits_rev(data, res + i, 8, 4 + bit_info + (i * 8));
		break;
		
		default:
		return 0;
	}
	*message = res;
	return 1;
}

static int QRData_make_final_data(QRData * data, int nb_bits)
{
	int bit_size;
	int i;
	int bit_info;
	bit_info = info_bits_size(data->mode);
	bit_size = 4 + bit_info + nb_bits;
	data->final_data_size = (bit_size / 8) + (bit_size % 8 != 0);
	/*printf("%d => %d\n", nb_bits, data->final_data_size);*/
	if((data->final_data = (unsigned char *)calloc(data->final_data_size + 1, sizeof(unsigned char))) == NULL)
		return 0;
	set_bits_rev(data->final_data, &(data->mode), 4, 0);
	set_bits_rev(data->final_data, &(data->infos), bit_info, 4);
	for(i = 0; i < nb_bits; i += 8)
	{
		/*printf("%d/%d (%d/%d)\n", i, nb_bits, 4 + bit_info + i, data->final_data_size * 8);*/
		set_bits_rev(data->final_data, data->data + (i / 8), 8, 4 + bit_info + i);
	}
	/*printf("%d, %d, %d\n", i, 8 + nb_bits - i, 4 + bit_info + i);*/
	set_bits_rev(data->final_data, data->data + (i / 8), 8 + nb_bits - i, 4 + bit_info + i);
	return 1;
}

int QRData_create_numeric(QRData * data, const char * numbers)
{
	int size;
	int len;
	int i;
	int j;
	int tmp;
	
	for(len = 0; *(numbers + len) != '\0'; ++len)
	{
		if(*(numbers + len) < '0' || *(numbers + len) > '9')
			return 0;
	}
	size = ((10 * len) / (3 * 8)) + (((10 * len) % (3 * 8)) != 0);
	if(! QRData_create(data, QRD_NUMERIC, size))
		return 0;
	data->infos = (unsigned int)len;
	for(i = 0; i + 3 <= len; i += 3)
	{
		tmp = 0;
		for(j = 0; j < 3; ++j)
			tmp += ((*(numbers + i + j) - '0') * power_int(10, 2 - j));
		set_bits_rev(data->data, &tmp, 10, 10 * (i / 3));
	}
	if(len - i != 0)
	{
		tmp = 0;
		for(j = 0; j < (len - i); ++j)
			tmp += ((*(numbers + i + j) - '0') * power_int(10, len - i - j - 1));
		set_bits_rev(data->data, &tmp, 1 + 3 * (len - i), 10 * (i / 3));
	}
	if(! QRData_make_final_data(data, ((len / 3) * 10) + (len - i > 0) * ((len - i) * 3 + 1)))
		return 0;
	return 1;
}

static int get_alphanumeric(int car)
{
	int i;
	const char * chaine = "$%*+-./:";
	if(car >= '0' && car <= '9')
		return car - '0';
	if(car >= 'A' && car <= 'Z')
		return car - 'A' + 10;
	if(car == ' ' || car == '\t' || car == '\n')
		return 36;
	for(i = 0; i < 8; ++i)
	{
		if(*(chaine + i) == car)
			return i + 37;
	}
	return -1;
}

int QRData_create_alphanumeric(QRData * data, const char * text)
{
	int size;
	int len;
	int i, j;
	int tmp;
	
	for(len = 0; *(text + len) != '\0'; ++len)
	{
		if(get_alphanumeric(*(text + len)) < 0)
			return 0;
	}
	size = ((11 * len) / (2 * 8)) + (((11 * len) % (2 * 8)) != 0);
	if(! QRData_create(data, QRD_ALPHANUMERIC, size))
		return 0;
	data->infos = (unsigned int)len;
	for(i = 0; (i / 2) < (len / 2); i += 2)
	{
		tmp = 0;
		for(j = 0; j < 2; ++j)
			tmp = tmp * 45 + get_alphanumeric(*(text + i + j));
		set_bits_rev(data->data, &tmp, 11, 11 * (i / 2));
	}
	if(len % 2 == 1)
	{
		tmp = get_alphanumeric(*(text + (len - 1)));
		set_bits_rev(data->data, &tmp, 6, 11 * (len / 2));
	}
	if(! QRData_make_final_data(data, ((len / 2) * 11) + (len % 2) * 6))
		return 0;
	return 1;
}

int QRData_create_byte(QRData * data, const char * bit_seq, int nb_bytes)
{
	int i;
	if(! QRData_create(data, QRD_BYTE, nb_bytes))
		return 0;
	data->infos = (unsigned int)nb_bytes;
	for(i = 0; i < nb_bytes; ++i)
		*(data->data + i) = *(bit_seq + i);
	if(! QRData_make_final_data(data, nb_bytes * 8))
		return 0;
	return 1;
}

static unsigned char * sized_seq(const unsigned char * seq, int seq_size, int final_size)
{
	unsigned char * res = NULL;
	int i;
	/*printf("%d - %d\n", seq_size, final_size);*/
	if(final_size < seq_size)
		return NULL;
	if((res = (unsigned char *)calloc(final_size, sizeof(unsigned char))) == NULL)
		return NULL;
	for(i = 0; i < seq_size; ++i)
		*(res + i) = *(seq + i);
	return res;
}

int QRData_make_code(QRData * data, QRCode * code, int version, CorrectionLevel cl, int mask_id)
{
	unsigned char * seq = NULL;
	if(! QRCode_create_version(code, version))
		return 0;
	if(! QRCode_set_format(code, cl, mask_id)
	|| ! QRCode_print_format(code))
	{
		QRCode_free(code);
		fprintf(stderr, "Erreur écriture format\n");
		return 0;
	}
	if((seq = sized_seq(data->final_data, data->final_data_size, get_capacity(version, cl))) == NULL)
	{
		QRCode_free(code);
		fprintf(stderr, "Erreur création de la séquence à coder\n");
		return 0;
	}
	if(! QRCode_code_data(code, seq))
	{
		free(seq);
		QRCode_free(code);
		fprintf(stderr, "Erreur codage\n");
		return 0;
	}
	free(seq);
	/*QRCode_save_autozoom(code, "media/sans_masque.bmp");*/
	QRCode_apply_mask(code);
	QRCode_print_format(code);
	return 1;
}

int QRData_create_text(QRData * data, const char * text)
{
	int can_be_numeric = 1;
	int can_be_alphanumeric = 1;
	int len;
	for(len = 0; *(text + len) != '\0'; ++len)
	{
		if(*(text + len) < '0' || *(text + len) > '9')
			can_be_numeric = 0;
		if(get_alphanumeric(*(text + len)) < 0)
			can_be_alphanumeric = 0;
	}
	if(can_be_numeric)
		return QRData_create_numeric(data, text);
	if(can_be_alphanumeric)
		return QRData_create_alphanumeric(data, text);
	return QRData_create_byte(data, text, len);
}

void QRData_print(const QRData * data)
{
	int i;
	printf("mode : %d\n", data->mode);
	printf("infos : %x\n", data->infos);
	printf("data : ");
	if(data->data == NULL)
	{
		printf("None\n");
		return;
	}
	for(i = 0; i < data->data_size; ++i)
		printf("%x ", *(data->data + i));
	printf("\n");
	printf("final data : ");
	if(data->final_data == NULL)
	{
		printf("None\n");
		return;
	}
	for(i = 0; i < data->final_data_size; ++i)
		printf("%x ", *(data->final_data + i));
	printf("\n");
}